#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Parses the Mbed configuration system and generates a CMake config script."""
import pathlib

from typing import Any

from mbed_tools.lib.json_helpers import decode_json_file
from mbed_tools.project import MbedProgram
from mbed_tools.targets import get_target_by_name
from mbed_tools.build._internal.cmake_file import render_mbed_config_cmake_template
from mbed_tools.build._internal.config.assemble_build_config import assemble_config
from mbed_tools.build._internal.write_files import write_file
from mbed_tools.build.exceptions import MbedBuildError

CMAKE_CONFIG_FILE = "mbed_config.cmake"


def generate_config(target_name: str, toolchain: str, program: MbedProgram) -> pathlib.Path:
    """Generate an Mbed config file at the program root by parsing the mbed config system.

    Args:
        target_name: Name of the target to configure for.
        toolchain: Name of the toolchain to use.
        program: The MbedProgram to configure.

    Returns:
        Path to the generated config file.
    """
    targets_data = _load_raw_targets_data(program)
    target_build_attributes = get_target_by_name(target_name, targets_data)
    config = assemble_config(
        target_build_attributes, [program.root, program.mbed_os.root], program.files.app_config_file
    )
    cmake_file_contents = render_mbed_config_cmake_template(
        target_name=target_name, config=config, toolchain_name=toolchain,
    )
    cmake_config_file_path = program.files.cmake_build_dir / CMAKE_CONFIG_FILE
    write_file(cmake_config_file_path, cmake_file_contents)
    return cmake_config_file_path


def _load_raw_targets_data(program: MbedProgram) -> Any:
    targets_data = decode_json_file(program.mbed_os.targets_json_file)
    if program.files.custom_targets_json.exists():
        custom_targets_data = decode_json_file(program.files.custom_targets_json)
        for custom_target in custom_targets_data:
            if custom_target in targets_data:
                raise MbedBuildError(
                    f"Error found in {program.files.custom_targets_json}.\n"
                    f"A target with the name '{custom_target}' already exists in targets.json. "
                    "Please give your custom target a unique name so it can be identified."
                )

        targets_data.update(custom_targets_data)

    return targets_data



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from tests.devices.markers import windows_only
from mbed_tools.devices._internal.windows.windows_identifier import WindowsUID

MOCKED_SERIAL_NUMBER_DATA = {
    "USB\\VID_04CA&PID_7058\\5&31AC2C0B&0&8": "6&212cca94&0",
    "USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900": None,
    "USB\\VID_1FD2&PID_5003\\5&31AC2C0B&0&10": "6&1f9e0013&0",
    "USB\\VID_0D28&PID_0204\\0240000034544E45001A00018AA900292011000097969900": None,
    "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&3": "7&3885ae75&0",
    "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4": "7&25cbfdd0&0",
    "USB\\ROOT_HUB30\\4&38EF038C&0&0": "5&31ac2c0b&0",
    "USB\\VID_2109&PID_2812\\5&31AC2C0B&0&1": "6&38e4ccb6&0",
    "USB\\VID_1366&PID_1015\\000440112138": "8&2f125ec6&0",
    "USB\\VID_1366&PID_1015&MI_02\\8&2F125EC6&0&0002": None,
    "USB\\VID_1366&PID_1015&MI_03\\8&2F125EC6&0&0003": "9&dbdecf6&0",
    "USB\\VID_1366&PID_1015\\000440109371": "8&37e0b92a&0",
    "USB\\VID_1366&PID_1015&MI_02\\8&37E0B92A&0&0002": None,
    "USB\\VID_1366&PID_1015&MI_03\\8&37E0B92A&0&0003": "9&4aa5a31&0",
    "USB\\VID_0483&PID_374B\\0670FF303931594E43184021": "8&2ae96f5b&0",
    "USB\\VID_0483&PID_374B&MI_01\\8&2AE96F5B&0&0001": "9&4a48a72&0",
}


def generate_mocked_system_usb_device_information():
    from mbed_tools.devices._internal.windows.usb_hub_data_loader import (
        SystemUsbDeviceInformation,
        UsbHub,
        UsbIdentifier,
    )
    from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader

    controllers = [
        UsbIdentifier(
            UID=WindowsUID(uid="3&33fd14ca&0", raw_uid="3&33FD14CA&0&A0", serial_number=None),
            VID="0483",
            PID="ABDB",
            REV="21",
            MI=None,
        )
    ]

    genuine_usb_hubs = [
        {"DeviceID": "USB\\VID_04CA&PID_7058\\5&31AC2C0B&0&8", "PNPDeviceID": "USB\\VID_04CA&PID_7058\\5&31AC2C0B&0&8"},
        {
            "DeviceID": "USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900",
            "PNPDeviceID": "USB\\VID_0D28&PID_0204&MI_00\\0240000034544E45001A00018AA900292011000097969900",
        },
        {
            "DeviceID": "USB\\VID_1FD2&PID_5003\\5&31AC2C0B&0&10",
            "PNPDeviceID": "USB\\VID_1FD2&PID_5003\\5&31AC2C0B&0&10",
        },
        {
            "DeviceID": "USB\\VID_0D28&PID_0204\\0240000034544E45001A00018AA900292011000097969900",
            "PNPDeviceID": "USB\\VID_0D28&PID_0204\\0240000034544E45001A00018AA900292011000097969900",
        },
        {"DeviceID": "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&3", "PNPDeviceID": "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&3"},
        {"DeviceID": "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4", "PNPDeviceID": "USB\\VID_2109&PID_2812\\6&38E4CCB6&0&4"},
        {"DeviceID": "USB\\ROOT_HUB30\\4&38EF038C&0&0", "PNPDeviceID": "USB\\ROOT_HUB30\\4&38EF038C&0&0"},
        {"DeviceID": "USB\\VID_2109&PID_2812\\5&31AC2C0B&0&1", "PNPDeviceID": "USB\\VID_2109&PID_2812\\5&31AC2C0B&0&1"},
        {"DeviceID": "USB\\VID_1366&PID_1015\\000440112138", "PNPDeviceID": "USB\\VID_1366&PID_1015\\000440112138"},
        {
            "DeviceID": "USB\\VID_1366&PID_1015&MI_02\\8&2F125EC6&0&0002",
            "PNPDeviceID": "USB\\VID_1366&PID_1015&MI_02\\8&2F125EC6&0&0002",
        },
        {
            "DeviceID": "USB\\VID_1366&PID_1015&MI_03\\8&2F125EC6&0&0003",
            "PNPDeviceID": "USB\\VID_1366&PID_1015&MI_03\\8&2F125EC6&0&0003",
        },
        {"DeviceID": "USB\\VID_1366&PID_1015\\000440109371", "PNPDeviceID": "USB\\VID_1366&PID_1015\\000440109371"},
        {
            "DeviceID": "USB\\VID_1366&PID_1015&MI_02\\8&37E0B92A&0&0002",
            "PNPDeviceID": "USB\\VID_1366&PID_1015&MI_02\\8&37E0B92A&0&0002",
        },
        {
            "DeviceID": "USB\\VID_1366&PID_1015&MI_03\\8&37E0B92A&0&0003",
            "PNPDeviceID": "USB\\VID_1366&PID_1015&MI_03\\8&37E0B92A&0&0003",
        },
        {
            "DeviceID": "USB\\VID_0483&PID_374B\\0670FF303931594E43184021",
            "PNPDeviceID": "USB\\VID_0483&PID_374B\\0670FF303931594E43184021",
        },
        {
            "DeviceID": "USB\\VID_0483&PID_374B&MI_01\\8&2AE96F5B&0&0001",
            "PNPDeviceID": "USB\\VID_0483&PID_374B&MI_01\\8&2AE96F5B&0&0001",
        },
    ]
    usb_hubs = genuine_usb_hubs + [
        dict(
            DeviceID=f"USB\\VID_{uid.vendor_id}&PID_{uid.product_id}&MI_01\\{uid.uid.raw_uid}",
            PNPDeviceID=f"USB\\VID_{uid.vendor_id}&PID_{uid.product_id}&MI_01\\{uid.uid.raw_uid}",
        )
        for uid in controllers
    ]

    class MockedDataLoader(SystemDataLoader):
        def _load(self):
            pass

    class MockedSystemUsbDeviceInformation(SystemUsbDeviceInformation):
        def __init__(self):
            super().__init__(MockedDataLoader())

        def _list_usb_controller_ids(self):
            return controllers

        def _iterate_over_hubs(self):
            hubs = [UsbHub() for _ in range(0, len(usb_hubs))]
            for i, hub in enumerate(usb_hubs):
                hubs[i].set_data_values(hub)
            for h in hubs:
                yield h

        def _determine_potential_serial_number(self, usb_device):
            return MOCKED_SERIAL_NUMBER_DATA.get(usb_device.pnp_id)

    return MockedSystemUsbDeviceInformation()


@windows_only
class TestUsbHub(unittest.TestCase):
    def test_system_usb_ids_list(self):
        from mbed_tools.devices._internal.windows.usb_hub_data_loader import UsbIdentifier

        mock = generate_mocked_system_usb_device_information()
        expected_values = {
            UsbIdentifier(
                UID=WindowsUID(
                    uid="0240000034544e45001a00018aa900292011000097969900",
                    raw_uid="0240000034544E45001A00018AA900292011000097969900",
                    serial_number=None,
                ),
                VID="0D28",
                PID="0204",
                REV=None,
                MI="00",
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="5&31ac2c0b&0", raw_uid="5&31AC2C0B&0&1", serial_number="6&38e4ccb6&0"),
                VID="2109",
                PID="2812",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="000440112138", raw_uid="000440112138", serial_number="8&2f125ec6&0"),
                VID="1366",
                PID="1015",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="4&38ef038c&0", raw_uid="4&38EF038C&0&0", serial_number="5&31ac2c0b&0"),
                VID=None,
                PID=None,
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="5&31ac2c0b&0", raw_uid="5&31AC2C0B&0&10", serial_number="6&1f9e0013&0"),
                VID="1FD2",
                PID="5003",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="000440109371", raw_uid="000440109371", serial_number="8&37e0b92a&0"),
                VID="1366",
                PID="1015",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="5&31ac2c0b&0", raw_uid="5&31AC2C0B&0&8", serial_number="6&212cca94&0"),
                VID="04CA",
                PID="7058",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(
                    uid="0670ff303931594e43184021", raw_uid="0670FF303931594E43184021", serial_number="8&2ae96f5b&0"
                ),
                VID="0483",
                PID="374B",
                REV=None,
                MI=None,
            ),
            UsbIdentifier(
                UID=WindowsUID(uid="6&38e4ccb6&0", raw_uid="6&38E4CCB6&0&3", serial_number="7&3885ae75&0"),
                VID="2109",
                PID="2812",
                REV=None,
                MI=None,
            ),
        }
        self.assertSetEqual(mock.usb_device_ids(), expected_values)

    def test_system_usb_interfaces(self):
        mock = generate_mocked_system_usb_device_information()
        cache = mock.usb_devices
        for id in mock.usb_device_ids():
            self.assertIn(id, cache)

    def test_get_usb_interfaces(self):
        from mbed_tools.devices._internal.windows.usb_hub_data_loader import UsbIdentifier

        known_usb = UsbIdentifier(
            UID=WindowsUID(
                uid="0670ff303931594e43184021", raw_uid="0670FF303931594E43184021", serial_number="8&2ae96f5b&0"
            ),
            VID="0483",
            PID="374B",
            REV=None,
            MI=None,
        )
        expected_related_interfaces = [
            "USB\\VID_0483&PID_374B\\0670FF303931594E43184021",
            "USB\\VID_0483&PID_374B&MI_01\\8&2AE96F5B&0&0001",
        ]
        mock = generate_mocked_system_usb_device_information()
        self.assertIsNotNone(mock.get_usb_devices(known_usb))
        self.assertListEqual([h.component_id for h in mock.get_usb_devices(known_usb)], expected_related_interfaces)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Public exceptions exposed by the package."""

from mbed_tools.lib.exceptions import ToolsError


class MbedProjectError(ToolsError):
    """Base exception for mbed-project."""


class VersionControlError(MbedProjectError):
    """Raised when a source control management operation failed."""


class ExistingProgram(MbedProjectError):
    """Raised when a program already exists at a given path."""


class ProgramNotFound(MbedProjectError):
    """Raised when an expected program is not found."""


class MbedOSNotFound(MbedProjectError):
    """A valid copy of MbedOS was not found."""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import mock

import sys

import pytest

from mbed_tools.sterm import terminal


@pytest.fixture
def mock_serial():
    with mock.patch("mbed_tools.sterm.terminal.Serial") as serial:
        yield serial


@pytest.fixture
def mock_console():
    with mock.patch("serial.tools.miniterm.Console") as console:
        yield console


@pytest.fixture
def mock_sterm():
    with mock.patch("mbed_tools.sterm.terminal.SerialTerminal") as sterm:
        yield sterm


def test_initialises_serial_port(mock_sterm, mock_serial):
    port = "tty.1111"
    baud = 9600

    terminal.run(port, baud)

    mock_serial.assert_called_once_with(port=port, baudrate=str(baud))


def test_initialises_sterm(mock_sterm, mock_serial):
    port = "tty.1111"
    baud = "9600"

    terminal.run(port, baud)

    mock_sterm.assert_called_once_with(mock_serial(), echo=True)


def test_starts_sterm_thread(mock_sterm, mock_serial):
    terminal.run("tty.122", 9600)

    mock_sterm().start.assert_called_once()


def test_joins_tx_and_rx_threads(mock_sterm, mock_serial):
    terminal.run("tty.122", 9600)

    mock_sterm().join.assert_any_call(True)


def test_joins_tx_thread_after_keyboard_interrupt(mock_sterm, mock_serial):
    mock_sterm().join.side_effect = (KeyboardInterrupt(), None)

    terminal.run("tty.122", 9600)

    mock_sterm().join.assert_called_with()


def test_closes_sterm(mock_sterm, mock_serial):
    terminal.run("tty.122", 9600)

    mock_sterm().close.assert_called_once()


def test_closes_sterm_after_exception(mock_sterm, mock_serial):
    mock_sterm().join.side_effect = (Exception(), None)
    with pytest.raises(Exception):
        terminal.run("tty.122", 9600)

    mock_sterm().close.assert_called_once()


def test_closes_sterm_after_keyboard_interrupt(mock_sterm, mock_serial):
    mock_sterm().join.side_effect = (KeyboardInterrupt(), None)
    terminal.run("tty.122", 9600)

    mock_sterm().close.assert_called_once()


def test_sets_terminal_special_chars(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())

    assert term.exit_character == terminal.CTRL_C
    assert term.menu_character == terminal.CTRL_T
    assert term.reset_character == terminal.CTRL_B
    assert term.help_character == terminal.CTRL_H


def test_sets_terminal_rx_and_tx_encoding_to_utf8(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())

    assert term.input_encoding == "UTF-8"
    assert term.output_encoding == "UTF-8"


def test_stops_terminal_when_ctrl_c_received(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    mock_console().getkey.return_value = terminal.CTRL_C

    term.writer()

    assert term.alive is False


def test_stops_terminal_on_keyboard_interrupt(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    mock_console().getkey.side_effect = KeyboardInterrupt()

    term.writer()

    assert term.alive is False


@pytest.mark.parametrize("menu_key", terminal.VALID_MENU_KEYS)
def test_handles_valid_menu_key(menu_key, mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.handle_menu_key = mock.Mock()
    term.alive = True
    mock_console().getkey.side_effect = (terminal.CTRL_T, menu_key, terminal.CTRL_C)

    term.writer()

    term.handle_menu_key.assert_called_once_with(menu_key)


INVALID_MENU_KEYS = tuple(set(chr(i) for i in range(0, 127)) - set(terminal.VALID_MENU_KEYS) - set([terminal.CTRL_H]))


@pytest.mark.parametrize("menu_key", INVALID_MENU_KEYS)
def test_ignores_invalid_menu_key(menu_key, mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.handle_menu_key = mock.Mock()
    term.alive = True
    mock_console().getkey.side_effect = (terminal.CTRL_T, menu_key)

    with pytest.raises(StopIteration):
        term.writer()

    term.handle_menu_key.assert_not_called()


def test_reset_sends_serial_break(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())

    term.reset()

    mock_serial().sendBreak.assert_called_once()


def test_ctrl_b_sends_reset_to_serial_port(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    mock_console().getkey.side_effect = (terminal.CTRL_B,)

    with pytest.raises(StopIteration):
        term.writer()

    mock_serial().sendBreak.assert_called_once()


def test_ctrl_h_prints_help_text(mock_serial, mock_console):
    sys.stderr.write = mock.Mock()
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    mock_console().getkey.side_effect = (terminal.CTRL_H,)

    with pytest.raises(StopIteration):
        term.writer()

    sys.stderr.write.assert_called_once_with(term.get_help_text())


def test_help_text_is_correct(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())

    assert term.get_help_text() == terminal.HELP_TEXT


def test_writes_normal_char_to_serial_output(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    normal_char = "h"
    mock_console().getkey.side_effect = (normal_char,)

    with pytest.raises(StopIteration):
        term.writer()

    mock_serial().write.assert_called_once_with(term.tx_encoder.encode(normal_char))


def test_echo_to_console_is_default_disabled(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial())
    term.alive = True
    normal_char = "h"
    mock_console().getkey.side_effect = (normal_char,)

    with pytest.raises(StopIteration):
        term.writer()

    mock_console().write.assert_not_called()


def test_echo_to_console_can_be_enabled(mock_serial, mock_console):
    term = terminal.SerialTerminal(mock_serial(), echo=True)
    term.alive = True
    normal_char = "h"
    mock_console().getkey.side_effect = (normal_char,)

    with pytest.raises(StopIteration):
        term.writer()

    mock_console().write.assert_called_once_with(normal_char)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import mock

import pytest

from click.testing import CliRunner

from mbed_tools.cli.sterm import sterm
from mbed_tools.devices.exceptions import MbedDevicesError


@pytest.fixture
def mock_terminal():
    with mock.patch("mbed_tools.cli.sterm.terminal") as term:
        yield term


@pytest.fixture
def mock_get_devices():
    with mock.patch("mbed_tools.cli.sterm.get_connected_devices") as get_devs:
        yield get_devs


@pytest.fixture
def mock_find_device():
    with mock.patch("mbed_tools.cli.sterm.find_connected_device") as find_dev:
        yield find_dev


def test_launches_terminal_on_given_serial_port(mock_terminal):
    port = "tty.1111"
    CliRunner().invoke(sterm, ["--port", port])

    mock_terminal.run.assert_called_once_with(port, 9600, echo=True)


def test_launches_terminal_with_given_baud_rate(mock_terminal):
    port = "tty.1111"
    baud = 115200
    CliRunner().invoke(sterm, ["--port", port, "--baudrate", baud])

    mock_terminal.run.assert_called_once_with(port, baud, echo=True)


def test_launches_terminal_with_echo_off_when_specified(mock_terminal):
    port = "tty.1111"
    CliRunner().invoke(sterm, ["--port", port, "--echo", "off"])

    mock_terminal.run.assert_called_once_with(port, 9600, echo=False)


def test_attempts_to_detect_device_if_no_port_given(mock_get_devices, mock_terminal):
    CliRunner().invoke(sterm, [])

    mock_get_devices.assert_called_once()


def test_attempts_to_find_connected_target_if_target_given(mock_find_device, mock_terminal):
    expected_port = "tty.k64f"
    mock_find_device.return_value = mock.Mock(serial_port=expected_port, mbed_board=mock.Mock(board_type="K64F"))

    CliRunner().invoke(sterm, ["-m", "K64F"])

    mock_terminal.run.assert_called_once_with(expected_port, 9600, echo=True)


def test_returns_serial_port_for_first_device_detected_if_no_target_given(mock_get_devices, mock_terminal):
    expected_port = "tty.k64f"
    mock_get_devices.return_value = mock.Mock(
        identified_devices=[
            mock.Mock(serial_port=expected_port, mbed_board=mock.Mock(board_type="K64F")),
            mock.Mock(serial_port="tty.disco", mbed_board=mock.Mock(board_type="DISCO")),
        ]
    )

    CliRunner().invoke(sterm, [])

    mock_terminal.run.assert_called_once_with(expected_port, 9600, echo=True)


def test_returns_serial_port_for_device_if_identifier_given(mock_find_device, mock_terminal):
    expected_port = "tty.k64f"
    mock_find_device.return_value = mock.Mock(serial_port=expected_port, mbed_board=mock.Mock(board_type="K64F"))

    CliRunner().invoke(sterm, ["-m", "K64F[1]"])

    mock_terminal.run.assert_called_once_with(expected_port, 9600, echo=True)


def test_raises_when_fails_to_find_default_device(mock_get_devices, mock_terminal):
    mock_get_devices.return_value = mock.Mock(identified_devices=[])

    with pytest.raises(MbedDevicesError):
        CliRunner().invoke(sterm, [], catch_exceptions=False)


def test_not_run_if_target_identifier_not_int(mock_get_devices, mock_terminal):
    target = "K64F[foo]"
    CliRunner().invoke(sterm, ["-m", target], catch_exceptions=False)
    mock_get_devices.assert_not_called()
    mock_terminal.assert_not_called()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Aggregation of all USB data given by Windows in various locations."""
from typing import NamedTuple, List, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor
from mbed_tools.devices._internal.windows.disk_aggregation import SystemDiskInformation, AggregatedDiskData
from mbed_tools.devices._internal.windows.serial_port import SerialPort
from mbed_tools.devices._internal.windows.serial_port_data_loader import SystemSerialPortInformation
from mbed_tools.devices._internal.windows.usb_device_identifier import UsbIdentifier
from mbed_tools.devices._internal.windows.usb_hub import UsbHub
from mbed_tools.devices._internal.windows.usb_hub_data_loader import SystemUsbDeviceInformation
from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader


class AggregatedUsbDataDefinition(NamedTuple):
    """Data aggregation with regards to a usb device."""

    usb_identifier: UsbIdentifier
    disks: List[AggregatedDiskData]
    serial_port: List[SerialPort]
    # On Windows, each interface e.g. Composite, Mass storage, Port is defined as
    # a separate independent UsbHub although they are related to the same device.
    related_usb_interfaces: List[UsbHub]


class AggregatedUsbData(ComponentDescriptor):
    """Usb information based on lots of different sources."""

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(AggregatedUsbDataDefinition, win32_class_name="AggregatedUsbData")

    @property
    def component_id(self) -> str:
        """Returns an id."""
        return str(self.uid)

    @property
    def uid(self) -> UsbIdentifier:
        """Returns the USB identifier."""
        return cast(UsbIdentifier, self.get("usb_identifier"))

    @property
    def is_associated_with_disk(self) -> bool:
        """States whether the usb device is associated with a disk."""
        return len(self.get("disks")) > 0

    @property
    def is_composite(self) -> bool:
        """States whether the usb device is associated with multiple interfaces."""
        return len(self.get("related_usb_interfaces")) > 1


class UsbDataAggregator:
    """Aggregator of all data related to a USB device."""

    def __init__(
        self,
        disk_data: SystemDiskInformation,
        serial_data: SystemSerialPortInformation,
        usb_data: SystemUsbDeviceInformation,
    ) -> None:
        """Initialiser."""
        self._disk_data = disk_data
        self._serial_data = serial_data
        self._usb_devices = usb_data

    def aggregate(self, usb_id: UsbIdentifier) -> AggregatedUsbData:
        """Aggregates data about a USB device from different sources."""
        disk_data = self._disk_data.get_disk_information(usb_id.uid)
        serial_data = self._serial_data.get_serial_port_information(usb_id)
        usb_data = self._usb_devices.get_usb_devices(usb_id)
        aggregated_data = AggregatedUsbData()
        aggregated_data.set_data_values(
            dict(usb_identifier=usb_id, disks=disk_data, serial_port=serial_data, related_usb_interfaces=usb_data,)
        )
        return aggregated_data


class SystemUsbData:
    """System in charge of gathering all the data related to USB devices."""

    def __init__(self, data_loader: SystemDataLoader) -> None:
        """Initialiser."""
        self._usb_devices = SystemUsbDeviceInformation(data_loader)
        self._aggregator = UsbDataAggregator(
            disk_data=SystemDiskInformation(data_loader),
            serial_data=SystemSerialPortInformation(data_loader),
            usb_data=self._usb_devices,
        )

    def all(self) -> List[AggregatedUsbData]:
        """Gets all the system data about USB devices."""
        return [self._aggregator.aggregate(usb_id) for usb_id in self._usb_devices.usb_device_ids()]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import plistlib
from unittest import TestCase, mock

from mbed_tools.devices._internal.darwin.diskutil import (
    get_all_external_disks_data,
    get_all_external_volumes_data,
    get_mount_point,
)


class TestGetAllExternalDisksData(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.diskutil.subprocess.check_output")
    def test_returns_disk_data_from_diskutil_call(self, check_output):
        check_output.return_value = b"""<plist version="1.0">
            <dict>
                <key>AllDisksAndPartitions</key>
                <array>
                    <string>le_data</string>
                </array>
            </dict>
        </plist>"""

        self.assertEqual(get_all_external_disks_data(), ["le_data"])


class TestGetAllExternalVolumesData(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.diskutil.get_all_external_disks_data")
    def test_returns_information_about_end_devices(self, get_all_external_disks_data):
        plist = b"""<plist version="1.0">
            <array>
                <dict>
                    <key>DeviceIdentifier</key>
                    <string>disk2</string>
                    <key>MountPoint</key>
                    <string>/Volumes/Foo</string>
                </dict>
                <dict>
                    <key>DeviceIdentifier</key>
                    <string>disk3</string>
                    <key>Partitions</key>
                    <array>
                        <dict>
                                <key>DeviceIdentifier</key>
                                <string>disk3s1</string>
                                <key>VolumeName</key>
                                <string>EFI</string>
                        </dict>
                        <dict>
                                <key>DeviceIdentifier</key>
                                <string>disk3s2</string>
                                <key>MountPoint</key>
                                <string>/Volumes/Untitled1</string>
                        </dict>
                        <dict>
                                <key>DeviceIdentifier</key>
                                <string>disk3s3</string>
                                <key>MountPoint</key>
                                <string>/Volumes/Untitled2</string>
                        </dict>
                    </array>
                </dict>
            </array>
        </plist>"""
        get_all_external_disks_data.return_value = plistlib.loads(plist)

        self.assertEqual(
            get_all_external_volumes_data(),
            [
                {"DeviceIdentifier": "disk2", "MountPoint": "/Volumes/Foo"},
                {"DeviceIdentifier": "disk3s1", "VolumeName": "EFI"},
                {"DeviceIdentifier": "disk3s2", "MountPoint": "/Volumes/Untitled1"},
                {"DeviceIdentifier": "disk3s3", "MountPoint": "/Volumes/Untitled2"},
            ],
        )


class TestGetMountPoint(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.diskutil.get_all_external_volumes_data")
    def test_returns_mountpoint_if_avaiable(self, get_all_external_volumes_data):
        get_all_external_volumes_data.return_value = [{"DeviceIdentifier": "disk123", "MountPoint": "/here"}]

        self.assertEqual(get_mount_point("disk123"), "/here")

    @mock.patch("mbed_tools.devices._internal.darwin.diskutil.get_all_external_volumes_data")
    def test_returns_none_if_no_mountpoint(self, get_all_external_volumes_data):
        get_all_external_volumes_data.return_value = []

        self.assertIsNone(get_mount_point("disk123"), None)

    @mock.patch("mbed_tools.devices._internal.darwin.diskutil.get_all_external_volumes_data")
    def test_handles_partial_data(self, get_all_external_volumes_data):
        get_all_external_volumes_data.return_value = [{"DeviceIdentifier": "disk4"}]

        self.assertIsNone(get_mount_point("disk4"), None)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from unittest.mock import patch

from tests.devices.markers import windows_only


@windows_only
class TestSystemDataLoader(unittest.TestCase):
    @patch("mbed_tools.devices._internal.windows.system_data_loader.load_all")
    def test_system_data_load(self, load_all):
        from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader, SYSTEM_DATA_TYPES

        def mock_system_element_fetcher(arg):
            return (arg, list())

        load_all.side_effect = mock_system_element_fetcher

        loader = SystemDataLoader()
        for type in SYSTEM_DATA_TYPES:
            self.assertIsNotNone(loader.get_system_data(type))
            self.assertTrue(isinstance(loader.get_system_data(type), list))
        load_all.assert_called()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from typing import NamedTuple, Any
from unittest import TestCase

from mbed_tools.devices._internal.windows.component_descriptor_utils import is_undefined_data_object, is_undefined_value
from tests.devices._internal.windows.test_component_descriptor_utils import (
    generate_undefined_values,
    generate_valid_values,
)
from tests.devices.markers import windows_only


class ComponentDefinition(NamedTuple):
    field1: str
    field2: str
    field3: str
    field4: Any
    field5: bool
    field6: int


def get_test_class():
    from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor

    class AComponentForTest(ComponentDescriptor):
        def __init__(self) -> None:
            """Initialiser."""
            super().__init__(ComponentDefinition, win32_class_name="Win32_ComputerSystem")

        @property
        def component_id(self) -> str:
            """Returns the device id field."""
            return self.get("field1")

    return AComponentForTest


@windows_only
class TestComponentDescriptor(TestCase):
    def test_init(self):
        self.assertIsNotNone(get_test_class()())
        self.assertTrue(get_test_class()().__class__, "AComponentForTest")

    def test_parameters(self):
        self.assertListEqual([name for name in ComponentDefinition._fields], get_test_class()().field_names)

    def test_set_values(self):

        a_component = get_test_class()()
        self.assertTrue(is_undefined_data_object(a_component.to_tuple()))
        valid_values = {k: generate_valid_values() for k in a_component.field_names}
        a_component.set_data_values(valid_values)
        self.assertFalse(is_undefined_data_object(a_component.to_tuple()))
        self.assertTupleEqual(a_component.to_tuple(), tuple(valid_values.values()))

    def test_is_undefined(self):
        a_component = get_test_class()()
        self.assertTrue(a_component.is_undefined)
        self.assertTrue(is_undefined_value(a_component.component_id))
        a_component_with_undefined_values = get_test_class()()
        undefined_values = {k: generate_undefined_values() for k in a_component.field_names}
        a_component_with_undefined_values.set_data_values(undefined_values)
        self.assertTrue(a_component_with_undefined_values.is_undefined)
        a_defined_component = get_test_class()()
        valid_values = {k: generate_valid_values() for k in a_component.field_names}
        a_defined_component.set_data_values(valid_values)
        self.assertIsNotNone(a_defined_component.component_id)
        self.assertFalse(a_defined_component.is_undefined)

    def test_iterator(self):
        from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptorWrapper

        generator = ComponentDescriptorWrapper(get_test_class()).element_generator()
        self.assertIsNotNone(generator)

        for element in generator:
            # The generator should be defined but none of the element should be defined as fields should not exist
            self.assertTrue(element.is_undefined)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Windows identifier."""
from functools import total_ordering
from typing import Any, Optional

from mbed_tools.devices._internal.windows.component_descriptor_utils import is_undefined_value

INSTANCE_ID_CHARACTER = "&"


def is_device_instance_id(value: Optional[str]) -> bool:
    """Determines whether a value is a device instance ID generated by Windows or not.

    See https://docs.microsoft.com/it-it/windows-hardware/drivers/install/instance-ids
    Typical IDs generated by Windows for a same device (but different interfaces) are, as follows:
        - 8&2F125EC6&0&0003
        - 8&2f125ec6&0
        - 8&2F125EC6&0&0002
    """
    return is_undefined_value(value) or INSTANCE_ID_CHARACTER in str(value)


@total_ordering
class WindowsUID:
    """Definition of a Windows Universal ID.

    UID in Windows can be either a serial number or an device instance ID depending
    on how underlying interfaces were defined.
    This object tries to hide the complexity of deciding which of the fields should be
    considered as the UID and defining
    equality between UIDs.
    Indeed different components referring to a same device may have completely different UIDs :
    e.g. 8&2F125EC6&0&0002, 8&2F125EC6&0&0003, 8&2f125ec6&0 and 345240562
    may all refer to the same component.
    """

    uid: str  # Processed string which may be the closest to the Universal ID.
    # This could be for the same device:
    # - 345240562
    # - 8&2F125EC6&0
    # - 8&2F125EC6
    # - 8&2f125ec6
    raw_uid: str  # The String from which the UID was derived.
    # Raw form, as it can be a combination of other elements
    # e.g.  8&2F125EC6&0&0003. In this case, the 0003 suffix corresponds to the MI value
    # and has nothing to do with the instance ID per se.
    serial_number: Optional[str]  # a serial number.

    # A string which can be the serial number of the device defined by
    # the underlying bus which may uniquely define the device.
    # e.g. 345240562. However, for some devices,
    # the serial number may not be accessible or defined or may actually be an instance ID.

    def __init__(self, uid: str, raw_uid: str, serial_number: Optional[str]) -> None:
        """Initialiser."""
        self.uid = uid
        self.raw_uid = raw_uid
        self.serial_number = serial_number

    def __eq__(self, other: Any) -> bool:
        """Defines the equality checker.

        The `equal` method does not follow a typical data model equal logic
        because of the complexity of this class which tries to find out which of its
        components is actually relevant and hence matters in determining equality.
        e.g.
        (uid="8&2f125ec6&0", raw_uid="8&2F125EC6&0&0003",serial_number="123456789")
        is the same as (uid="123456789", raw_uid="123456789",serial_number="8&2f125ec6")
        or (uid="8&2f125ec6&0", raw_uid="8&2F125EC6&0&000",serial_number=None)
        """
        if not other or not isinstance(other, WindowsUID):
            return False
        if (
            self.uid == other.uid
            or self.raw_uid == other.raw_uid
            or self.serial_number == other.serial_number
            or self.serial_number == other.uid
            or self.uid == other.serial_number
        ):
            return True
        # Due to the complexity of determining the UID on Windows,
        # we can assume that we are dealing with the same ID if IDs are subsets of each other.
        return str(self.uid).startswith(str(other.uid)) or str(other.uid).startswith(str(self.uid))

    @property
    def presumed_serial_number(self) -> str:
        """Determines what may be the most likely value for the serial number of the component.

        From the different components at its disposal, the system tries to find what may be the serial number.
        """
        if not is_device_instance_id(self.uid):
            # the UID is not an instance ID and should therefore be the serial number.
            return self.uid
        return self.uid if is_device_instance_id(self.serial_number) else str(self.serial_number)

    @property
    def instance_id(self) -> str:
        """Determines what may be the value most likely to be the instance_id of a component as generated by Windows."""
        if is_device_instance_id(self.uid):
            return self.uid
        return self.uid if is_undefined_value(self.serial_number) else str(self.serial_number)

    def contains_genuine_serial_number(self) -> bool:
        """Contains a genuine serial number and not an instance ID."""
        serial_number = self.presumed_serial_number
        return not (is_undefined_value(serial_number) or is_device_instance_id(serial_number))

    def __hash__(self) -> int:
        """Calculates the hash of the UID.

        Due to the complexity of the `equal` method, this implementation of the `hash`
        calculation breaks the hashing/equality rules of typical data objects.
        In the present case, the WindowsUID is mostly used for dictionary lookup and therefore,
        the actual use cases were tested.
        In most cases, the instance ID of a device will be provided rather than the serial number and therefore,
        this field will be considered as differentiators between UIDs.
        """
        return hash(self.instance_id)

    def __repr__(self) -> str:
        """String representation of the UID."""
        return f"WindowsUID({self.uid})"

    def __str__(self) -> str:
        """String representation of the UID."""
        elements = [f"{k}={v!r}" for (k, v) in self.__dict__.items()]
        return f"WindowsUID({', '.join(elements)})"

    def __lt__(self, other: "WindowsUID") -> bool:
        """Defines less than."""
        return self.presumed_serial_number < other.presumed_serial_number



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from unittest import TestCase

from mbed_tools.devices._internal.htm_file import read_product_code, read_online_id, OnlineId


class TestReadProductCode(TestCase):
    def test_reads_product_code_from_code_attribute(self):
        code = "02400201B80ECE4A45F033F2"
        file_contents = f'<meta http-equiv="refresh" content="0; url=http://mbed.org/device/?code={code}"/>'

        self.assertEqual(read_product_code(file_contents), code[:4])

    def test_reads_product_code_from_auth_attribute(self):
        auth = "101000000000000000000002F7F35E602eeb0bb9b632205c51f6c357aeee7bc9"
        file_contents = (
            '<meta http-equiv="refresh" '
            f'content="0; url=http://mbed.org/start?auth={auth}&loader=11972&firmware=16457&configuration=4" />'
        )

        self.assertEqual(read_product_code(file_contents), auth[:4])

    def test_none_if_no_product_code(self):
        self.assertIsNone(read_product_code(""))


class TestReadOnlineId(TestCase):
    def test_reads_online_id_from_url(self):
        url = "https://os.mbed.com/platforms/THIS-IS_a_SLUG_123/"
        file_contents = f"window.location.replace({url});"

        self.assertEqual(read_online_id(file_contents), OnlineId(target_type="platform", slug="THIS-IS_a_SLUG_123"))

    def test_none_if_not_found(self):
        file_contents = "window.location.replace(https://os.mbed.com/about);"
        self.assertIsNone(read_online_id(file_contents))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import json
from pathlib import Path
from tempfile import TemporaryDirectory

from mbed_tools.build._internal.config.assemble_build_config import _assemble_config_from_sources
from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.find_files import find_files
from mbed_tools.build._internal.config.source import prepare
from mbed_tools.lib.json_helpers import decode_json_file


def create_files(directory, files):
    created_files = []
    for file in files:
        path = Path(directory, file["path"])
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text(json.dumps(file["json_contents"]))
        created_files.append(path)
    return created_files


class TestAssembleConfigFromSourcesAndLibFiles:
    def test_assembles_config_using_all_relevant_files(self):
        target = {
            "config": {"foo": {"value": None}},
            "macros": [],
            "labels": ["A"],
            "extra_labels": [],
            "features": ["RED"],
            "components": [],
            "c_lib": "std",
            "printf_lib": "minimal-printf",
        }
        mbed_lib_files = [
            {
                "path": Path("subdir", "FEATURE_RED", "mbed_lib.json"),
                "json_contents": {
                    "name": "red",
                    "config": {"bool": {"value": False}},
                    "target_overrides": {
                        "A": {"bool": True, "target.features_add": ["BLUE"], "target.components_add": ["LEG"]}
                    },
                    "macros": ["RED_MACRO"],
                },
            },
            {
                "path": Path("TARGET_A", "mbed_lib.json"),
                "json_contents": {
                    "name": "a",
                    "config": {"number": {"value": 123}},
                    "target_overrides": {"*": {"target.features_add": ["RED"]}},
                },
            },
            {
                "path": Path("COMPONENT_LEG", "mbed_lib.json"),
                "json_contents": {
                    "name": "leg",
                    "config": {"number-of-fingers": {"value": 5}},
                    "macros": ["LEG_MACRO"],
                },
            },
        ]
        unused_mbed_lib_file = {
            "path": Path("subdir", "FEATURE_BROWN", "mbed_lib.json"),
            "json_contents": {
                "name": "brown",
                "target_overrides": {"*": {"red.bool": "DON'T USE ME"}},
                "macros": ["DONT_USE_THIS_MACRO"],
            },
        }
        mbed_app_file = {
            "path": Path("mbed_app.json"),
            "json_contents": {"target_overrides": {"*": {"target.foo": "bar"}}},
        }

        with TemporaryDirectory() as directory:
            created_mbed_lib_files = create_files(directory, mbed_lib_files)
            created_mbed_app_file = create_files(directory, [mbed_app_file])[0]
            create_files(directory, [unused_mbed_lib_file])

            subject = _assemble_config_from_sources(
                target, find_files("mbed_lib.json", Path(directory)), created_mbed_app_file
            )

            mbed_lib_sources = [
                prepare(decode_json_file(Path(directory, file)), target_filters=["A"])
                for file in created_mbed_lib_files
            ]
            mbed_app_source = prepare(decode_json_file(Path(directory, created_mbed_app_file)), target_filters=["A"])
            expected_config = Config(prepare(target, source_name="target"))
            for source in mbed_lib_sources + [mbed_app_source]:
                expected_config.update(source)

            subject["config"].sort(key=lambda x: x.name)
            expected_config["config"].sort(key=lambda x: x.name)
            assert subject == expected_config

    def test_updates_target_labels_from_config(self):
        target = {
            "config": {"foo": None},
            "macros": {"NO"},
            "labels": {"A"},
            "extra_labels": set(),
            "features": set(),
            "components": set(),
            "c_lib": "std",
            "printf_lib": "minimal-printf",
        }
        mbed_lib_files = [
            {
                "path": Path("TARGET_A", "mbed_lib.json"),
                "json_contents": {"name": "a", "target_overrides": {"*": {"target.features_add": ["RED"]}}},
            },
            {
                "path": Path("subdir", "FEATURE_RED", "mbed_lib.json"),
                "json_contents": {
                    "name": "red",
                    "target_overrides": {
                        "A": {
                            "target.macros_remove": ["NO"],
                            "target.features_add": ["BLUE"],
                            "target.components_add": ["LEG"],
                        }
                    },
                    "macros": ["RED_MACRO"],
                },
            },
        ]
        mbed_app_file = {
            "path": Path("mbed_app.json"),
            "json_contents": {
                "target_overrides": {
                    "*": {
                        "target.foo": "bar",
                        "target.labels_add": ["PICKLE"],
                        "target.extra_labels_add": ["EXTRA_HOT"],
                        "target.macros_add": ["TICKER"],
                    }
                }
            },
        }

        with TemporaryDirectory() as directory:
            _ = create_files(directory, mbed_lib_files)
            created_mbed_app_file = create_files(directory, [mbed_app_file])[0]

            config = _assemble_config_from_sources(
                target, find_files("mbed_lib.json", Path(directory)), created_mbed_app_file
            )

            assert config["features"] == {"RED", "BLUE"}
            assert config["components"] == {"LEG"}
            assert config["extra_labels"] == {"EXTRA_HOT"}
            assert config["labels"] == {"A", "PICKLE"}
            assert config["macros"] == {"TICKER", "RED_MACRO"}



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Command to build/compile an Mbed project using CMake."""
import os
import pathlib
import shutil

from typing import Optional, Tuple

import click

from mbed_tools.build import build_project, generate_build_system, generate_config, flash_binary
from mbed_tools.devices import find_connected_device, find_all_connected_devices
from mbed_tools.project import MbedProgram
from mbed_tools.sterm import terminal


@click.command(name="compile", help="Build an Mbed project.")
@click.option(
    "-t",
    "--toolchain",
    type=click.Choice(["ARM", "GCC_ARM"], case_sensitive=False),
    help="The toolchain you are using to build your app.",
)
@click.option("-m", "--mbed-target", help="A build target for an Mbed-enabled device, e.g. K64F.")
@click.option("-b", "--profile", default="develop", help="The build type (release, develop or debug).")
@click.option("-c", "--clean", is_flag=True, default=False, help="Perform a clean build.")
@click.option(
    "-p",
    "--program-path",
    default=os.getcwd(),
    help="Path to local Mbed program. By default it is the current working directory.",
)
@click.option(
    "--mbed-os-path", type=click.Path(), default=None, help="Path to local Mbed OS directory.",
)
@click.option(
    "--custom-targets-json", type=click.Path(), default=None, help="Path to custom_targets.json.",
)
@click.option(
    "-f", "--flash", is_flag=True, default=False, help="Flash the binary onto a device",
)
@click.option(
    "--hex-file", is_flag=True, default=False, help="Use hex file, this option should be used with '-f/--flash' option",
)
@click.option(
    "-s", "--sterm", is_flag=True, default=False, help="Launch a serial terminal to the device.",
)
@click.option(
    "--baudrate",
    default=9600,
    show_default=True,
    help="Change the serial baud rate (ignored unless --sterm is also given).",
)
def build(
    program_path: str,
    profile: str,
    toolchain: str = "",
    mbed_target: str = "",
    clean: bool = False,
    flash: bool = False,
    hex_file: bool = False,
    sterm: bool = False,
    baudrate: int = 9600,
    mbed_os_path: str = None,
    custom_targets_json: str = None,
) -> None:
    """Configure and build an Mbed project using CMake and Ninja.

    If the CMake configuration step has already been run previously (i.e a CMake build tree exists), then just try to
    build the project immediately using Ninja.

    Args:
       program_path: Path to the Mbed project.
       mbed_os_path: The path to the local Mbed OS directory.
       profile: The Mbed build profile (debug, develop or release).
       custom_targets_json: Path to custom_targets.json.
       toolchain: The toolchain to use for the build.
       mbed_target: The name of the Mbed target to build for.
       clean: Perform a clean build.
       flash: Flash the binary onto a device.
       hex_file: Use hex file, this option should be used with '-f/--flash' option.
       sterm: Open a serial terminal to the connected target.
       baudrate: Change the serial baud rate (ignored unless --sterm is also given).
    """
    _validate_target_and_toolchain_args(mbed_target, toolchain)
    mbed_target, target_id = _get_target_id(mbed_target)

    cmake_build_subdir = pathlib.Path(mbed_target.upper(), profile.lower(), toolchain.upper())
    if mbed_os_path is None:
        program = MbedProgram.from_existing(pathlib.Path(program_path), cmake_build_subdir)
    else:
        program = MbedProgram.from_existing(pathlib.Path(program_path), cmake_build_subdir, pathlib.Path(mbed_os_path))
    build_tree = program.files.cmake_build_dir
    if clean and build_tree.exists():
        shutil.rmtree(build_tree)

    click.echo("Configuring project and generating build system...")
    if custom_targets_json is not None:
        program.files.custom_targets_json = pathlib.Path(custom_targets_json)
    generate_config(mbed_target.upper(), toolchain, program)
    generate_build_system(program.root, build_tree, profile)

    click.echo("Building Mbed project...")
    build_project(build_tree)

    if flash or sterm:
        if target_id is not None or sterm:
            devices = [find_connected_device(mbed_target, target_id)]
        else:
            devices = find_all_connected_devices(mbed_target)

    if flash:
        for dev in devices:
            flashed_path = flash_binary(dev.mount_points[0].resolve(), program.root, build_tree, mbed_target, hex_file)
        click.echo(f"Copied {str(flashed_path.resolve())} to {len(devices)} device(s).")
    elif hex_file:
        click.echo("'--hex-file' option should be used with '-f/--flash' option")

    if sterm:
        dev = devices[0]
        if dev.serial_port is None:
            raise click.ClickException(
                f"The connected device {dev.mbed_board.board_name} does not have an associated serial port."
                " Reconnect the device and try again."
            )

        terminal.run(dev.serial_port, baudrate)


def _validate_target_and_toolchain_args(target: str, toolchain: str) -> None:
    if not all([toolchain, target]):
        raise click.UsageError(
            "Both --toolchain and --mbed-target arguments are required when using the compile subcommand."
        )


def _get_target_id(target: str) -> Tuple[str, Optional[int]]:
    if "[" in target:
        target_name, target_id = target.replace("]", "").split("[", maxsplit=1)
        if target_id.isdigit() and int(target_id) >= 0:
            return (target_name, int(target_id))
        raise click.ClickException("When using the format mbed-target[ID], ID must be a positive integer or 0.")
    return (target, None)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets._internal.board_database`."""

from unittest import TestCase, mock

import requests_mock

# Unit under test
import mbed_tools.targets._internal.board_database as board_database
from mbed_tools.targets.env import env


class TestGetOnlineBoardData(TestCase):
    """Tests for the method `board_database.get_online_board_data`."""

    @requests_mock.mock()
    @mock.patch("mbed_tools.targets._internal.board_database.logger.warning", autospec=True)
    @mock.patch("mbed_tools.targets._internal.board_database.logger.debug", autospec=True)
    def test_401(self, mock_request, logger_debug, logger_warning):
        """Given a 401 error code, BoardAPIError is raised."""
        mock_request.get(board_database._BOARD_API, status_code=401, text="Who are you?")
        with self.assertRaises(board_database.BoardAPIError):
            board_database.get_online_board_data()
        self.assertTrue("MBED_API_AUTH_TOKEN" in str(logger_warning.call_args), "Auth token should be mentioned")
        self.assertTrue("Who are you?" in str(logger_debug.call_args), "Message content should be in the debug message")

    @requests_mock.mock()
    @mock.patch("mbed_tools.targets._internal.board_database.logger.warning", autospec=True)
    @mock.patch("mbed_tools.targets._internal.board_database.logger.debug", autospec=True)
    def test_404(self, mock_request, logger_debug, logger_warning):
        """Given a 404 error code, TargetAPIError is raised."""
        mock_request.get(board_database._BOARD_API, status_code=404, text="Not Found")
        with self.assertRaises(board_database.BoardAPIError):
            board_database.get_online_board_data()
        self.assertTrue("404" in str(logger_warning.call_args), "HTTP status code should be mentioned")
        self.assertTrue("Not Found" in str(logger_debug.call_args), "Message content should be in the debug message")

    @requests_mock.mock()
    @mock.patch("mbed_tools.targets._internal.board_database.logger.warning", autospec=True)
    @mock.patch("mbed_tools.targets._internal.board_database.logger.debug", autospec=True)
    def test_200_invalid_json(self, mock_request, logger_debug, logger_warning):
        """Given a valid response but invalid json, JSONDecodeError is raised."""
        mock_request.get(board_database._BOARD_API, text="some text")
        with self.assertRaises(board_database.ResponseJSONError):
            board_database.get_online_board_data()
        self.assertTrue("Invalid JSON" in str(logger_warning.call_args), "Invalid JSON should be mentioned")
        self.assertTrue("some text" in str(logger_debug.call_args), "Message content should be in the debug message")

    @requests_mock.mock()
    @mock.patch("mbed_tools.targets._internal.board_database.logger.warning", autospec=True)
    @mock.patch("mbed_tools.targets._internal.board_database.logger.debug", autospec=True)
    def test_200_no_data_field(self, mock_request, logger_debug, logger_warning):
        """Given a valid response but no data field, ResponseJSONError is raised."""
        mock_request.get(board_database._BOARD_API, json={"notdata": [], "stillnotdata": {}})
        with self.assertRaises(board_database.ResponseJSONError):
            board_database.get_online_board_data()
        self.assertTrue("missing the 'data' field" in str(logger_warning.call_args), "Data field should be mentioned")
        self.assertTrue(
            "notdata, stillnotdata" in str(logger_debug.call_args),
            "JSON keys from message should be in the debug message",
        )

    @requests_mock.mock()
    def test_200_value_data(self, mock_request):
        """Given a valid response, target data is set from the returned json."""
        mock_request.get(board_database._BOARD_API, json={"data": 42})
        board_data = board_database.get_online_board_data()
        self.assertEqual(42, board_data, "Target data should match the contents of the target API data")

    @mock.patch("mbed_tools.targets._internal.board_database.requests")
    @mock.patch("mbed_tools.targets._internal.board_database.env", spec_set=env)
    def test_auth_header_set_with_token(self, env, requests):
        """Given an authorization token env variable, get is called with authorization header."""
        env.MBED_API_AUTH_TOKEN = "token"
        header = {"Authorization": "Bearer token"}
        board_database._get_request()
        requests.get.assert_called_once_with(board_database._BOARD_API, headers=header)

    @mock.patch("mbed_tools.targets._internal.board_database.requests")
    def test_no_auth_header_set_with_empty_token_var(self, requests):
        """Given no authorization token env variable, get is called with no header."""
        board_database._get_request()
        requests.get.assert_called_once_with(board_database._BOARD_API, headers=None)

    @mock.patch("mbed_tools.targets._internal.board_database.requests.get")
    def test_raises_tools_error_on_connection_error(self, get):
        get.side_effect = board_database.requests.exceptions.ConnectionError
        with self.assertRaises(board_database.BoardAPIError):
            board_database._get_request()


class TestGetOfflineTargetData(TestCase):
    """Tests for the method get_offline_target_data."""

    def test_local_target_database_file_found(self):
        """Test local database is found and loaded."""
        data = board_database.get_offline_board_data()
        self.assertTrue(len(data), "Some data should be returned from local database file.")

    @mock.patch("mbed_tools.targets._internal.board_database.get_board_database_path")
    def test_raises_on_invalid_json(self, mocked_get_file):
        """Test raises an error when the file contains invalid JSON."""
        invalid_json = "None"
        path_mock = mock.Mock()
        path_mock.read_text.return_value = invalid_json
        mocked_get_file.return_value = path_mock
        with self.assertRaises(board_database.ResponseJSONError):
            board_database.get_offline_board_data()


class TestGetLocalTargetDatabaseFile(TestCase):
    def test_returns_path_to_targets(self):
        path = board_database.get_board_database_path()
        self.assertEqual(path.exists(), True, "Path to boards should exist in the package data folder.")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Loads serial port data."""
from typing import Optional, Generator, cast, List

from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader, ComponentsLoader
from mbed_tools.devices._internal.windows.usb_device_identifier import UsbIdentifier, parse_device_id
from mbed_tools.devices._internal.windows.serial_port import SerialPort


class SystemSerialPortInformation:
    """All information about the serial ports of the current system."""

    def __init__(self, data_loader: SystemDataLoader) -> None:
        """Initialiser."""
        self._serial_port_by_usb_id: Optional[dict] = None
        self._data_loader = data_loader

    def _load_data(self) -> None:
        self._serial_port_by_usb_id = {
            parse_device_id(p.pnp_id): p
            for p in cast(
                Generator[SerialPort, None, None], ComponentsLoader(self._data_loader, SerialPort).element_generator()
            )
        }

    @property
    def serial_port_data_by_id(self) -> dict:
        """Gets system's serial ports by usb id."""
        if not self._serial_port_by_usb_id:
            self._load_data()
        return self._serial_port_by_usb_id if self._serial_port_by_usb_id else dict()

    def get_serial_port_information(self, usb_id: UsbIdentifier) -> List[SerialPort]:
        """Gets all disk information for a given serial number."""
        port = self.serial_port_data_by_id.get(usb_id)
        return [port] if port else list()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Utility in charge of finding the instance ID of a device."""
import win32con
import win32api
from mbed_tools.devices._internal.exceptions import SystemException
import logging

from typing import Optional, Any

logger = logging.getLogger(__name__)


class RegKey(object):
    """Context manager in charge of opening and closing registry keys."""

    def __init__(self, sub_registry_key: str) -> None:
        """Initialiser."""
        access = win32con.KEY_READ | win32con.KEY_ENUMERATE_SUB_KEYS | win32con.KEY_QUERY_VALUE
        try:
            self._hkey = win32api.RegOpenKey(win32con.HKEY_LOCAL_MACHINE, sub_registry_key, 0, access)
        except win32api.error as e:
            raise SystemException(f"Could not read key [{sub_registry_key}] in the registry: {e}")

    def __enter__(self) -> Any:
        """Actions on entry."""
        return self._hkey

    def __exit__(self, type: Any, value: Any, traceback: Any) -> None:
        """Actions on exit."""
        win32api.RegCloseKey(self._hkey)
        self._hkey.close()


def get_children_instance_id(pnpid: str) -> Optional[str]:
    """Gets the USB children instance ID from the plug and play ID.

    See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/instance-ids.
    """
    # Although the registry should not be accessed directly
    # (See https://docs.microsoft.com/en-us/windows-hardware/drivers/install/hklm-system-currentcontrolset-enum-registry-tree), # noqa E501
    # and SetupDi functions/APIs should be used instead in a similar fashion to Miro utility
    # (See https://github.com/cool-RR/Miro/blob/7b9ecd9bc0878e463f5a5e26e8b00b675e3f98ac/tv/windows/plat/usbutils.py)
    # Most libraries seems to be reading the registry:
    #     - Pyserial: https://github.com/pyserial/pyserial/blob/master/serial/tools/list_ports_windows.py
    #     - Node serialport: https://github.com/serialport/node-serialport/blob/cd112ca5a3a3fe186e1ac6fa78eeeb5ea7396185/packages/bindings/src/serialport_win.cpp # noqa E501
    #     - USB device forensics: https://github.com/woanware/usbdeviceforensics/blob/master/pyTskusbdeviceforensics.py
    # For more details about the registry key actually looked at, See:
    # - https://stackoverflow.com/questions/3331043/get-list-of-connected-usb-devices
    # - https://docs.microsoft.com/en-us/windows-hardware/drivers/usbcon/usb-device-specific-registry-settings
    key_path = f"SYSTEM\\CurrentControlSet\\Enum\\{pnpid}"
    with RegKey(key_path) as hkey:
        try:
            value = win32api.RegQueryValueEx(hkey, "ParentIdPrefix")
            return str(value[0]) if value else None
        except win32api.error as e:
            logger.debug(f"Error occurred reading `ParentIdPrefix` field of key [{key_path}] in the registry: {e}")
            return None



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Creation and management of Mbed OS projects.

* Creation of a new Mbed OS application.
* Cloning of an existing Mbed OS program.
* Deploy of a specific version of Mbed OS or library.
"""

from mbed_tools.project.project import initialise_project, import_project, deploy_project, get_known_libs
from mbed_tools.project.mbed_program import MbedProgram



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from pathlib import Path
from unittest import mock

import pytest

from mbed_tools.project.exceptions import VersionControlError
from mbed_tools.project._internal import git_utils


@pytest.fixture
def mock_repo():
    with mock.patch("mbed_tools.project._internal.git_utils.git.Repo") as repo:
        yield repo


@pytest.fixture
def mock_progress():
    with mock.patch("mbed_tools.project._internal.git_utils.ProgressReporter") as progress:
        yield progress


class TestClone:
    def test_returns_repo(self, mock_progress, mock_repo, tmp_path):
        url = "https://blah"
        path = Path(tmp_path, "dst")
        repo = git_utils.clone(url, path)

        assert repo is not None
        mock_repo.clone_from.assert_called_once_with(url, str(path), progress=mock_progress())

    def test_raises_when_fails_due_to_bad_url(self, tmp_path):
        with pytest.raises(VersionControlError, match="from url 'bad' failed"):
            git_utils.clone("bad", Path(tmp_path, "dst"))

    def test_raises_when_fails_due_to_existing_nonempty_dst_dir(self, mock_repo, tmp_path):
        dst_dir = Path(tmp_path, "dst")
        dst_dir.mkdir()
        (dst_dir / "some_file.txt").touch()

        with pytest.raises(VersionControlError, match="exists and is not an empty directory"):
            git_utils.clone("https://blah", dst_dir)

    def test_can_clone_to_empty_existing_dst_dir(self, mock_repo, tmp_path, mock_progress):
        dst_dir = Path(tmp_path, "dst")
        dst_dir.mkdir()
        url = "https://repo"

        repo = git_utils.clone(url, dst_dir)

        assert repo is not None
        mock_repo.clone_from.assert_called_once_with(url, str(dst_dir), progress=mock_progress())


class TestInit:
    def test_returns_initialised_repo(self, mock_repo):
        repo = git_utils.init(Path())

        assert repo is not None
        mock_repo.init.assert_called_once_with(str(Path()))

    def test_raises_when_init_fails(self, mock_repo):
        mock_repo.init.side_effect = git_utils.git.exc.GitCommandError("git init", 255)

        with pytest.raises(VersionControlError):
            git_utils.init(Path())


class TestGetRepo:
    def test_returns_repo_object(self, mock_repo):
        repo = git_utils.get_repo(Path())

        assert isinstance(repo, mock_repo().__class__)

    def test_raises_version_control_error_when_no_git_repo_found(self, mock_repo):
        mock_repo.side_effect = git_utils.git.exc.InvalidGitRepositoryError

        with pytest.raises(VersionControlError):
            git_utils.get_repo(Path())


class TestCheckout:
    def test_git_lib_called_with_correct_command(self, mock_repo):
        git_utils.checkout(mock_repo, "master")

        mock_repo.git.checkout.assert_called_once_with("master")

    def test_git_lib_called_with_correct_command_with_force(self, mock_repo):
        git_utils.checkout(mock_repo, "master", force=True)

        mock_repo.git.checkout.assert_called_once_with("master", "--force")

    def test_raises_version_control_error_when_git_checkout_fails(self, mock_repo):
        mock_repo.git.checkout.side_effect = git_utils.git.exc.GitCommandError("git checkout", 255)

        with pytest.raises(VersionControlError):
            git_utils.checkout(mock_repo, "bad")


class TestGetDefaultBranch:
    def test_returns_default_branch_name(self, mock_repo):
        mock_repo().git.symbolic_ref.return_value = "refs/remotes/origin/main"

        branch_name = git_utils.get_default_branch(mock_repo())

        assert branch_name == "main"

    def test_raises_version_control_error_when_git_command_fails(self, mock_repo):
        mock_repo().git.symbolic_ref.side_effect = git_utils.git.exc.GitCommandError("git symbolic_ref", 255)

        with pytest.raises(VersionControlError):
            git_utils.get_default_branch(mock_repo())



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Configuration source parser."""
import logging
import pathlib

from dataclasses import dataclass
from typing import Iterable, Any, Optional, List

from mbed_tools.lib.json_helpers import decode_json_file
from mbed_tools.build.exceptions import InvalidConfigOverride
from mbed_tools.lib.python_helpers import flatten_nested

logger = logging.getLogger(__name__)


def from_file(
    config_source_file_path: pathlib.Path, target_filters: Iterable[str], default_name: Optional[str] = None
) -> dict:
    """Load a JSON config file and prepare the contents as a config source."""
    return prepare(decode_json_file(config_source_file_path), source_name=default_name, target_filters=target_filters)


def prepare(
    input_data: dict, source_name: Optional[str] = None, target_filters: Optional[Iterable[str]] = None
) -> dict:
    """Prepare a config source for entry into the Config object.

    Extracts config and override settings from the source. Flattens these nested dictionaries out into lists of
    objects which are namespaced in the way the Mbed config system expects.

    Args:
        input_data: The raw config JSON object parsed from the config file.
        source_name: Optional default name to use for namespacing config settings. If the input_data contains a 'name'
            field, that field is used as the namespace.
        target_filters: List of filter string used when extracting data from target_overrides section of the config
            data.

    Returns:
        Prepared config source.
    """
    data = input_data.copy()
    namespace = data.pop("name", source_name)
    for key in data:
        data[key] = _sanitise_value(data[key])

    if "config" in data:
        data["config"] = _extract_config_settings(namespace, data["config"])

    if "overrides" in data:
        data["overrides"] = _extract_overrides(namespace, data["overrides"])

    if "target_overrides" in data:
        data["overrides"] = _extract_target_overrides(
            namespace, data.pop("target_overrides"), target_filters if target_filters is not None else []
        )

    return data


@dataclass
class ConfigSetting:
    """Representation of a config setting.

    Auto converts any list values to sets for faster operations and de-duplication of values.
    """

    namespace: str
    name: str
    value: Any
    help_text: Optional[str] = None
    macro_name: Optional[str] = None

    def __post_init__(self) -> None:
        """Convert the value to a set if applicable."""
        self.value = _sanitise_value(self.value)


@dataclass
class Override:
    """Representation of a config override.

    Checks for _add or _remove modifiers and splits them from the name.
    """

    namespace: str
    name: str
    value: Any
    modifier: Optional[str] = None

    def __post_init__(self) -> None:
        """Parse modifiers and convert list values to sets."""
        if self.name.endswith("_add") or self.name.endswith("_remove"):
            self.name, self.modifier = self.name.rsplit("_", maxsplit=1)

        self.value = _sanitise_value(self.value)


def _extract_config_settings(namespace: str, config_data: dict) -> List[ConfigSetting]:
    settings = []
    for name, item in config_data.items():
        logger.debug("Extracting config setting '%s'='%s'", name, item)
        if isinstance(item, dict):
            macro_name = item.get("macro_name")
            help_text = item.get("help")
            value = item.get("value")
        else:
            macro_name = None
            help_text = None
            value = item

        setting = ConfigSetting(
            namespace=namespace, name=name, macro_name=macro_name, help_text=help_text, value=value,
        )
        # If the config item is about a certain component or feature
        # being present, avoid adding it to the mbed_config.cmake
        # configuration file. Instead, applications should depend on
        # the feature or component with target_link_libraries() and the
        # component's CMake file (in the Mbed OS repo) will create
        # any necessary macros or definitions.
        if setting.name == "present":
            continue

        settings.append(setting)

    return settings


def _extract_target_overrides(
    namespace: str, override_data: dict, allowed_target_labels: Iterable[str]
) -> List[Override]:
    valid_target_data = dict()
    for target_type in override_data:
        if target_type == "*" or target_type in allowed_target_labels:
            valid_target_data.update(override_data[target_type])

    return _extract_overrides(namespace, valid_target_data)


def _extract_overrides(namespace: str, override_data: dict) -> List[Override]:
    overrides = []
    for name, value in override_data.items():
        try:
            override_namespace, override_name = name.split(".")
            if override_namespace and override_namespace not in [namespace, "target"] and namespace != "app":
                raise InvalidConfigOverride(
                    "It is only possible to override config settings defined in an mbed_lib.json from mbed_app.json. "
                    f"An override was defined by the lib `{namespace}` that attempts to override "
                    f"`{override_namespace}.{override_name}`."
                )
        except ValueError:
            override_namespace = namespace
            override_name = name

        overrides.append(Override(namespace=override_namespace, name=override_name, value=value))

    return overrides


def _sanitise_value(val: Any) -> Any:
    """Convert list values to sets and return scalar values and strings unchanged.

    For whatever reason, we allowed config settings to have values of any type available in the JSON spec.
    The value type can be a list, nested list, str, int, you name it.

    When we process the config, we want to use sets instead of lists, this is for two reasons:
    * To take advantage of set operations when we deal with "cumulative" settings.
    * To prevent any duplicate settings ending up in the final config.
    """
    if isinstance(val, list):
        return set(flatten_nested(val))

    return val



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interface for accessing Mbed-Enabled Development Board data.

An instance of `mbed_tools.targets.board.Board` can be retrieved by calling one of the public functions.
"""
import logging
from enum import Enum
from typing import Callable

from mbed_tools.targets.env import env
from mbed_tools.targets.exceptions import UnknownBoard, UnsupportedMode, BoardDatabaseError
from mbed_tools.targets.board import Board
from mbed_tools.targets.boards import Boards


logger = logging.getLogger(__name__)


def get_board_by_product_code(product_code: str) -> Board:
    """Returns first `mbed_tools.targets.board.Board` matching given product code.

    Args:
        product_code: the product code to look up in the database.

    Raises:
        UnknownBoard: a board with a matching product code was not found.
    """
    return get_board(lambda board: board.product_code == product_code)


def get_board_by_online_id(slug: str, target_type: str) -> Board:
    """Returns first `mbed_tools.targets.board.Board` matching given online id.

    Args:
        slug: The slug to look up in the database.
        target_type: The target type to look up in the database, normally one of `platform` or `module`.

    Raises:
        UnknownBoard: a board with a matching slug and target type could not be found.
    """
    matched_slug = slug.casefold()
    return get_board(lambda board: board.slug.casefold() == matched_slug and board.target_type == target_type)


def get_board(matching: Callable) -> Board:
    """Returns first `mbed_tools.targets.board.Board` for which `matching` is True.

    Uses database mode configured in the environment.

    Args:
        matching: A function which will be called to test matching conditions for each board in database.

    Raises:
        UnknownBoard: a board matching the criteria could not be found in the board database.
    """
    database_mode = _get_database_mode()

    if database_mode == _DatabaseMode.OFFLINE:
        logger.info("Using the offline database (only) to identify boards.")
        return Boards.from_offline_database().get_board(matching)

    if database_mode == _DatabaseMode.ONLINE:
        logger.info("Using the online database (only) to identify boards.")
        return Boards.from_online_database().get_board(matching)
    try:
        logger.info("Using the offline database to identify boards.")
        return Boards.from_offline_database().get_board(matching)
    except UnknownBoard:
        logger.info("Unable to identify a board using the offline database, trying the online database.")
        try:
            return Boards.from_online_database().get_board(matching)
        except BoardDatabaseError:
            logger.error("Unable to access the online database to identify a board.")
            raise UnknownBoard()


class _DatabaseMode(Enum):
    """Selected database mode."""

    OFFLINE = 0
    ONLINE = 1
    AUTO = 2


def _get_database_mode() -> _DatabaseMode:
    database_mode = env.MBED_DATABASE_MODE
    try:
        return _DatabaseMode[database_mode]
    except KeyError:
        raise UnsupportedMode(f"{database_mode} is not a supported database mode.")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Interactions with `diskutil`."""
import plistlib
import subprocess
from typing import Dict, Iterable, List, Optional, cast
from typing_extensions import TypedDict


VolumeTree = Dict  # mypy does not work with recursive types, which nested "Partitions" would require


class Volume(TypedDict, total=False):
    """Representation of mounted volume."""

    MountPoint: str  # example: /Volumes/SomeName
    DeviceIdentifier: str  # example: disk2


def get_all_external_disks_data() -> List[VolumeTree]:
    """Returns parsed output of `diskutil` call, fetching only information of interest."""
    output = subprocess.check_output(["diskutil", "list", "-plist", "external"], stderr=subprocess.DEVNULL)
    if output:
        data: Dict = plistlib.loads(output)
        return data.get("AllDisksAndPartitions", [])
    return []


def get_all_external_volumes_data() -> List[Volume]:
    """Returns all external volumes data.

    Reduces structure returned by `diskutil` call to one which will only contain data about Volumes.
    Useful for determining MountPoints and DeviceIdentifiers.
    """
    data = get_all_external_disks_data()
    return _filter_volumes(data)


def get_external_volume_data(device_identifier: str) -> Optional[Volume]:
    """Returns external volume data for a given identifier."""
    data = get_all_external_volumes_data()
    for device in data:
        if device.get("DeviceIdentifier") == device_identifier:
            return device
    return None


def get_mount_point(device_identifier: str) -> Optional[str]:
    """Returns mount point of a given device."""
    device_data = get_external_volume_data(device_identifier)
    if device_data and "MountPoint" in device_data:
        return device_data["MountPoint"]
    return None


def _filter_volumes(data: Iterable[VolumeTree]) -> List[Volume]:
    """Flattens the structure returned by `diskutil` call.

    Expected input will contain both partitioned an unpartitioned devices.
    Partitioned devices list mounted partitions under an arbitrary key,
    flattening the data helps finding actual end devices later on.
    """
    devices = []
    for device in data:
        if "Partitions" in device:
            devices.extend(_filter_volumes(device["Partitions"]))
        else:
            devices.append(cast(Volume, device))
    return devices



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import subprocess

from unittest import mock

import pytest

from mbed_tools.build.build import build_project, generate_build_system
from mbed_tools.build.exceptions import MbedBuildError


@pytest.fixture
def subprocess_run():
    with mock.patch("mbed_tools.build.build.subprocess.run", autospec=True) as subproc:
        yield subproc


class TestBuildProject:
    def test_invokes_cmake_with_correct_args(self, subprocess_run):
        build_project(build_dir="cmake_build", target="install")

        subprocess_run.assert_called_with(["cmake", "--build", "cmake_build", "--target", "install"], check=True)

    def test_invokes_cmake_with_correct_args_if_no_target_passed(self, subprocess_run):
        build_project(build_dir="cmake_build")

        subprocess_run.assert_called_with(["cmake", "--build", "cmake_build"], check=True)

    def test_raises_build_error_if_cmake_invocation_fails(self, subprocess_run):
        subprocess_run.side_effect = (None, subprocess.CalledProcessError(1, ""))

        with pytest.raises(MbedBuildError, match="CMake invocation failed"):
            build_project(build_dir="cmake_build")


class TestConfigureProject:
    def test_invokes_cmake_with_correct_args(self, subprocess_run):
        source_dir = "source_dir"
        build_dir = "cmake_build"
        profile = "debug"

        generate_build_system(source_dir, build_dir, profile)

        subprocess_run.assert_called_with(
            ["cmake", "-S", source_dir, "-B", build_dir, "-GNinja", f"-DCMAKE_BUILD_TYPE={profile}"], check=True
        )

    def test_raises_when_ninja_cannot_be_found(self, subprocess_run):
        subprocess_run.side_effect = FileNotFoundError

        with pytest.raises(MbedBuildError, match="Ninja"):
            generate_build_system("", "", "")

    def test_raises_when_cmake_cannot_be_found(self, subprocess_run):
        subprocess_run.side_effect = (None, FileNotFoundError)

        with pytest.raises(MbedBuildError, match="Could not find CMake"):
            generate_build_system("", "", "")



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pytest

from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.config.source import prepare, ConfigSetting, Override


class TestConfig:
    def test_config_updated(self):
        conf = Config()

        conf.update(prepare({"config": {"param": {"value": 0}}}, source_name="lib"))
        conf.update(prepare({"config": {"param2": {"value": 0}}}, source_name="lib2"))

        assert conf["config"][0].name == "param"
        assert conf["config"][1].name == "param2"

    def test_raises_when_trying_to_add_duplicate_config_setting(self):
        conf = Config(prepare({"config": {"param": {"value": 0}}}, source_name="lib"))

        with pytest.raises(ValueError, match="lib.param already defined"):
            conf.update(prepare({"config": {"param": {"value": 0}}}, source_name="lib"))

    def test_target_overrides_handled(self):
        conf = Config(
            {
                "config": [
                    ConfigSetting(namespace="target", name="network-default-interface-type", help_text="", value="WIFI")
                ],
                "device_has": ["TEST"],
            }
        )

        conf.update(
            {
                "overrides": [
                    Override(namespace="target", name="network-default-interface-type", value="ETHERNET"),
                    Override(namespace="target", name="device_has", value={"OVERRIDDEN"}),
                ]
            }
        )

        network_iface, *_ = conf["config"]
        assert network_iface.value == "ETHERNET"
        assert conf["device_has"] == {"OVERRIDDEN"}

    def test_target_overrides_separate_namespace(self):
        conf = Config(
            {
                "config": [
                    ConfigSetting(
                        namespace="dontchangeme", name="network-default-interface-type", help_text="", value="WIFI"
                    ),
                    ConfigSetting(
                        namespace="changeme", name="network-default-interface-type", help_text="", value="WIFI"
                    ),
                ]
            }
        )

        conf.update(
            {"overrides": [Override(namespace="changeme", name="network-default-interface-type", value="ETHERNET")]}
        )

        dontchangeme, changeme, *_ = conf["config"]
        assert changeme.namespace == "changeme"
        assert changeme.value == "ETHERNET"
        assert dontchangeme.namespace == "dontchangeme"
        assert dontchangeme.value == "WIFI"

    def test_lib_overrides_handled(self):
        conf = Config(
            {
                "config": [
                    ConfigSetting(namespace="lib", name="network-default-interface-type", help_text="", value="WIFI")
                ],
            }
        )

        conf.update({"overrides": [Override(namespace="lib", name="network-default-interface-type", value="ETHERNET")]})

        network_iface, *_ = conf["config"]
        assert network_iface.value == "ETHERNET"

    def test_cumulative_fields_can_be_modified(self):
        conf = Config({"device_has": {"FLASHING_LIGHTS"}, "macros": {"A"}})

        conf.update(
            {
                "overrides": [
                    Override(namespace="target", name="device_has", modifier="add", value={"OTHER_STUFF"}),
                    Override(namespace="target", name="device_has", modifier="remove", value={"FLASHING_LIGHTS"}),
                    Override(namespace="lib", name="macros", modifier="remove", value={"A"}),
                    Override(namespace="target", name="macros", modifier="add", value={"B"}),
                ]
            }
        )

        conf.update({"overrides": [Override(namespace="target", name="macros", modifier="add", value={"B"})]})
        assert conf["device_has"] == {"OTHER_STUFF"}
        assert conf["macros"] == {"B"}

    def test_macros_are_appended_to(self):
        conf = Config({"macros": {"A"}})

        conf.update({"macros": {"B"}})
        conf.update({"macros": {"B"}})

        assert conf["macros"] == {"A", "B"}

    def test_warns_and_skips_override_for_undefined_config_parameter(self, caplog):
        conf = Config()
        override_name = "this-does-not-exist"
        conf.update(prepare({"target_overrides": {"*": {override_name: ""}}}))
        assert override_name in caplog.text
        assert not conf

    def test_ignores_present_option(self):
        source = prepare({"name": "mbed_component", "config": {"present": {"help": "Mbed Component", "value": True}}})

        config = Config(source)

        assert not config["config"]



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Disk partition."""

from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor


class DiskPartitionMsdnDefinition(NamedTuple):
    """Msdn definition of a disk partition.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-diskpartition
    """

    AdditionalAvailability: int
    Availability: int
    PowerManagementCapabilities: list
    IdentifyingDescriptions: list
    MaxQuiesceTime: int
    OtherIdentifyingInfo: int
    StatusInfo: int
    PowerOnHours: int
    TotalPowerOnHours: int
    Access: int
    BlockSize: int
    Bootable: bool
    BootPartition: bool
    Caption: str
    ConfigManagerErrorCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    Description: str
    DeviceID: str
    DiskIndex: int
    ErrorCleared: bool
    ErrorDescription: str
    ErrorMethodology: str
    HiddenSectors: int
    Index: int
    InstallDate: int
    LastErrorCode: int
    Name: str
    NumberOfBlocks: int
    PNPDeviceID: str
    PowerManagementSupported: bool
    PrimaryPartition: bool
    Purpose: str
    RewritePartition: bool
    Size: int
    StartingOffset: int
    Status: str
    SystemCreationClassName: str
    SystemName: str
    Type: str


class DiskPartition(ComponentDescriptor):
    """Disk partition as defined in Windows API.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/win32-diskpartition
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(DiskPartitionMsdnDefinition, win32_class_name="Win32_DiskPartition")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Internal helper to retrieve target attribute information.

This information is parsed from the targets.json configuration file
found in the mbed-os repo.
"""
import logging
import pathlib
from typing import Dict, Any, Set, Optional

from mbed_tools.lib.exceptions import ToolsError
from mbed_tools.lib.json_helpers import decode_json_file

from mbed_tools.targets._internal.targets_json_parsers.accumulating_attribute_parser import (
    get_accumulating_attributes_for_target,
)
from mbed_tools.targets._internal.targets_json_parsers.overriding_attribute_parser import (
    get_overriding_attributes_for_target,
    get_labels_for_target,
)

INTERNAL_PACKAGE_DIR = pathlib.Path(__file__).parent
MBED_OS_METADATA_FILE = pathlib.Path(INTERNAL_PACKAGE_DIR, "data", "targets_metadata.json")

logger = logging.getLogger(__name__)


class TargetAttributesError(ToolsError):
    """Target attributes error."""


class ParsingTargetsJSONError(TargetAttributesError):
    """targets.json parsing failed."""


class TargetNotFoundError(TargetAttributesError):
    """Target definition not found in targets.json."""


def get_target_attributes(targets_json_data: dict, target_name: str) -> dict:
    """Retrieves attribute data taken from targets.json for a single target.

    Args:
        targets_json_data: target definitions from targets.json
        target_name: the name of the target (often a Board's board_type).

    Returns:
        A dictionary representation of the attributes for the target.

    Raises:
        ParsingTargetJSONError: error parsing targets.json
        TargetNotFoundError: there is no target attribute data found for that target.
    """
    target_attributes = _extract_target_attributes(targets_json_data, target_name)
    target_attributes["labels"] = get_labels_for_target(targets_json_data, target_name).union(
        _extract_core_labels(target_attributes.get("core", None))
    )
    target_attributes["extra_labels"] = set(target_attributes.get("extra_labels", []))
    target_attributes["features"] = set(target_attributes.get("features", []))
    target_attributes["components"] = set(target_attributes.get("components", []))
    target_attributes["macros"] = set(target_attributes.get("macros", []))
    target_attributes["config"] = _apply_config_overrides(
        target_attributes.get("config", {}), target_attributes.get("overrides", {})
    )
    return target_attributes


def _extract_target_attributes(all_targets_data: Dict[str, Any], target_name: str) -> dict:
    """Extracts the definition for a particular target from all the targets in targets.json.

    Args:
        all_targets_data: a dictionary representation of the raw targets.json data.
        target_name: the name of the target.

    Returns:
        A dictionary representation the target definition.

    Raises:
        TargetNotFoundError: no target definition found in targets.json.
    """
    if target_name not in all_targets_data:
        raise TargetNotFoundError(f"Target attributes for {target_name} not found.")

    # All target definitions are assumed to be public unless specifically set as public=false
    if not all_targets_data[target_name].get("public", True):
        raise TargetNotFoundError(f"Target attributes for {target_name} not found.")

    target_attributes = get_overriding_attributes_for_target(all_targets_data, target_name)
    accumulated_attributes = get_accumulating_attributes_for_target(all_targets_data, target_name)
    target_attributes.update(accumulated_attributes)
    return target_attributes


def _extract_core_labels(target_core: Optional[str]) -> Set[str]:
    """Find the labels associated with the target's core.

    Args:
        target_core: the target core, set as a build attribute

    Returns:
        A list of labels associated with the target's core, or an empty set
        if either core is undefined or no labels found for the core.
    """
    if target_core:
        mbed_os_metadata = decode_json_file(MBED_OS_METADATA_FILE)
        return set(mbed_os_metadata["CORE_LABELS"].get(target_core, []))
    return set()


def _apply_config_overrides(config: Dict[str, Any], overrides: Dict[str, Any]) -> Dict[str, Any]:
    """Returns the config attribute with any overrides applied.

    Args:
        config: the cumulative config settings defined for a target
        overrides: the values that need to be changed in the config settings for this target

    Returns:
        The config settings with the overrides applied.

    Raises:
        TargetsJsonConfigurationError: overrides can't be applied to config settings that aren't already defined
    """
    config = config.copy()
    for key in overrides:
        try:
            config[key]["value"] = overrides[key]
        except KeyError:
            logger.warning(
                f"Cannot apply override {key}={overrides[key]}, there is no config setting defined matching that name."
            )
    return config



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Progress bar for git operations."""
import sys

from typing import Optional, Any

from git import RemoteProgress
from tqdm import tqdm


class ProgressBar(tqdm):
    """tqdm progress bar that can be used in a callback context."""

    total: Any

    def update_progress(self, block_num: float = 1, block_size: float = 1, total_size: float = None) -> None:
        """Update the progress bar.

        Args:
            block_num: Number of the current block.
            block_size: Size of the current block.
            total_size: Total size of all expected blocks.
        """
        if total_size is not None and self.total != total_size:
            self.total = total_size
        self.update(block_num * block_size - self.n)


class ProgressReporter(RemoteProgress):
    """GitPython RemoteProgress subclass that displays a progress bar for git fetch and push operations."""

    def __init__(self, *args: Any, name: str = "", **kwargs: Any) -> None:
        """Initialiser.

        Args:
            name: The name of the git repository to report progress on.
        """
        self.name = name
        super().__init__(*args, **kwargs)

    def update(self, op_code: int, cur_count: float, max_count: Optional[float] = None, message: str = "") -> None:
        """Called whenever the progress changes.

        Args:
            op_code: Integer describing the stage of the current operation.
            cur_count: Current item count.
            max_count: Maximum number of items expected.
            message: Message string describing the number of bytes transferred in the WRITING operation.
        """
        if self.BEGIN & op_code:
            self.bar = ProgressBar(total=max_count, file=sys.stderr, leave=False)

        self.bar.desc = f"{self.name} {self._cur_line}"
        self.bar.update_progress(block_num=cur_count, total_size=max_count)

        if self.END & op_code:
            self.bar.close()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import contextlib
import os
import pathlib

from tempfile import TemporaryDirectory
from unittest import TestCase, mock

from click.testing import CliRunner

from mbed_tools.cli.build import build
from mbed_tools.project._internal.project_data import BUILD_DIR
from mbed_tools.build.config import CMAKE_CONFIG_FILE


DEFAULT_BUILD_ARGS = ["-t", "GCC_ARM", "-m", "K64F"]
DEFAULT_BUILD_SUBDIR = pathlib.Path("K64F", "develop", "GCC_ARM")


@contextlib.contextmanager
def mock_project_directory(
    program, mbed_config_exists=False, build_tree_exists=False, build_subdir=DEFAULT_BUILD_SUBDIR
):
    with TemporaryDirectory() as tmp_dir:
        root = pathlib.Path(tmp_dir, "test-project")
        root.mkdir()
        program.root = root
        program.files.cmake_build_dir = root / BUILD_DIR / build_subdir
        program.files.cmake_config_file = root / BUILD_DIR / build_subdir / CMAKE_CONFIG_FILE
        if mbed_config_exists:
            program.files.cmake_config_file.parent.mkdir(exist_ok=True, parents=True)
            program.files.cmake_config_file.touch(exist_ok=True)

        if build_tree_exists:
            program.files.cmake_build_dir.mkdir(exist_ok=True, parents=True)

        yield


@mock.patch("mbed_tools.cli.build.generate_build_system")
@mock.patch("mbed_tools.cli.build.build_project")
@mock.patch("mbed_tools.cli.build.MbedProgram")
@mock.patch("mbed_tools.cli.build.generate_config")
class TestBuildCommand(TestCase):
    def test_searches_for_mbed_program_at_default_project_path(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        runner = CliRunner()
        runner.invoke(build, DEFAULT_BUILD_ARGS)

        mbed_program.from_existing.assert_called_once_with(pathlib.Path(os.getcwd()), DEFAULT_BUILD_SUBDIR)

    def test_searches_for_mbed_program_at_user_defined_project_root(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        project_path = "my-blinky"

        runner = CliRunner()
        runner.invoke(build, ["-p", project_path, *DEFAULT_BUILD_ARGS])

        mbed_program.from_existing.assert_called_once_with(pathlib.Path(project_path), DEFAULT_BUILD_SUBDIR)

    def test_calls_generate_build_system_if_build_tree_nonexistent(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=False):
            runner = CliRunner()
            runner.invoke(build, DEFAULT_BUILD_ARGS)

            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")

    def test_generate_config_called_if_config_script_nonexistent(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(
            program, mbed_config_exists=False, build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM")
        ):
            target = "k64f"
            toolchain = "gcc_arm"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)

    def test_raises_if_gen_config_toolchain_not_passed_when_required(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=False):
            target = "k64f"

            runner = CliRunner()
            result = runner.invoke(build, ["-m", target])

            self.assertIsNotNone(result.exception)
            self.assertRegex(result.output, "--toolchain")

    def test_raises_if_gen_config_target_not_passed_when_required(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=False):
            toolchain = "gcc_arm"

            runner = CliRunner()
            result = runner.invoke(build, ["-t", toolchain])

            self.assertIsNotNone(result.exception)
            self.assertRegex(result.output, "--mbed-target")

    def test_raises_if_gen_config_target_toolchain_not_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program):
            runner = CliRunner()
            result = runner.invoke(build)

            self.assertIsNotNone(result.exception)
            self.assertRegex(result.output, "--mbed-target")
            self.assertRegex(result.output, "--toolchain")

    def test_raises_if_target_identifier_not_int(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        target = "K64F[foo]"

        result = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm"])
        self.assertIsNotNone(result.exception)
        self.assertRegex(result.output, "ID")

    def test_raises_if_target_identifier_negative(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        target = "K64F[-1]"

        result = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm"])
        self.assertIsNotNone(result.exception)
        self.assertRegex(result.output, "ID")

    def test_build_system_regenerated_when_mbed_os_path_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(
            program,
            mbed_config_exists=True,
            build_tree_exists=True,
            build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM"),
        ):
            toolchain = "gcc_arm"
            target = "k64f"
            mbed_os_path = "./extern/mbed-os"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target, "--mbed-os-path", mbed_os_path])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")

    def test_custom_targets_location_used_when_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(program, mbed_config_exists=True, build_tree_exists=True):
            toolchain = "gcc_arm"
            target = "k64f"
            custom_targets_json_path = pathlib.Path("custom", "custom_targets.json")

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target, "--custom-targets-json", custom_targets_json_path])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
            self.assertEqual(program.files.custom_targets_json, custom_targets_json_path)

    def test_build_folder_removed_when_clean_flag_passed(
        self, generate_config, mbed_program, build_project, generate_build_system
    ):
        program = mbed_program.from_existing()
        with mock_project_directory(
            program,
            mbed_config_exists=True,
            build_tree_exists=True,
            build_subdir=pathlib.Path("K64F", "develop", "GCC_ARM"),
        ):
            toolchain = "gcc_arm"
            target = "k64f"

            runner = CliRunner()
            runner.invoke(build, ["-t", toolchain, "-m", target, "-c"])

            generate_config.assert_called_once_with(target.upper(), toolchain.upper(), program)
            generate_build_system.assert_called_once_with(program.root, program.files.cmake_build_dir, "develop")
            self.assertFalse(program.files.cmake_build_dir.exists())

    @mock.patch("mbed_tools.cli.build.flash_binary")
    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
    def test_build_flash_option(
        self, mock_find_devices, flash_binary, generate_config, mbed_program, build_project, generate_build_system
    ):
        mock_find_devices.return_value = [mock.MagicMock()]
        runner = CliRunner()
        runner.invoke(build, ["--flash", *DEFAULT_BUILD_ARGS])
        flash_binary.assert_called_once()

    @mock.patch("mbed_tools.cli.build.flash_binary")
    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
    def test_build_flash_and_hex_file_options(
        self, mock_find_devices, flash_binary, generate_config, mbed_program, build_project, generate_build_system
    ):
        mock_find_devices.return_value = [mock.MagicMock()]
        runner = CliRunner()
        runner.invoke(build, ["--flash", "--hex-file", *DEFAULT_BUILD_ARGS])
        call_args = flash_binary.call_args
        args, kwargs = call_args
        flash_binary.assert_called_once_with(args[0], args[1], args[2], args[3], True)

    @mock.patch("mbed_tools.cli.build.flash_binary")
    @mock.patch("mbed_tools.cli.build.find_all_connected_devices")
    def test_build_flash_both_two_devices(
        self, mock_find_devices, flash_binary, generate_config, mbed_program, build_project, generate_build_system
    ):
        mock_find_devices.return_value = [mock.MagicMock(), mock.MagicMock()]
        runner = CliRunner()
        runner.invoke(build, ["--flash", *DEFAULT_BUILD_ARGS])
        self.assertEqual(flash_binary.call_count, 2)

    @mock.patch("mbed_tools.cli.build.flash_binary")
    @mock.patch("mbed_tools.cli.build.find_connected_device")
    def test_build_flash_only_identifier_device(
        self, mock_find_device, flash_binary, generate_config, mbed_program, build_project, generate_build_system
    ):
        mock_find_device.return_value = mock.MagicMock()
        runner = CliRunner()
        runner.invoke(build, ["--flash", "-m", "K64F[1]", "-t", "GCC_ARM"])
        self.assertEqual(flash_binary.call_count, 1)

    def test_build_only_hex_file_option(self, generate_config, mbed_program, build_project, generate_build_system):
        runner = CliRunner()
        result = runner.invoke(build, ["--hex-file", *DEFAULT_BUILD_ARGS])

        self.assertRegex(result.output, "-f/--flash")

    @mock.patch("mbed_tools.cli.build.terminal")
    @mock.patch("mbed_tools.cli.build.find_connected_device")
    def test_sterm_is_started_when_flag_passed(
        self, mock_find_device, mock_terminal, generate_config, mbed_program, build_project, generate_build_system
    ):
        target = "K64F"
        serial_port = "tty.k64f"
        baud = 115200
        mock_find_device.return_value = mock.Mock(serial_port=serial_port)

        CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm", "--sterm", "--baudrate", baud])

        mock_find_device.assert_called_once_with(target, None)
        mock_terminal.run.assert_called_once_with(serial_port, baud)

    @mock.patch("mbed_tools.cli.build.terminal")
    @mock.patch("mbed_tools.cli.build.find_connected_device")
    def test_raises_if_device_does_not_have_serial_port_and_sterm_flag_given(
        self, mock_find_device, mock_terminal, generate_config, mbed_program, build_project, generate_build_system
    ):
        target = "K64F"
        serial_port = None
        mock_find_device.return_value = mock.Mock(serial_port=serial_port)

        output = CliRunner().invoke(build, ["-m", target, "-t", "gcc_arm", "--sterm"])
        self.assertEqual(type(output.exception), SystemExit)
        mock_terminal.assert_not_called()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Serial terminal implementation based on pyserial.tools.miniterm.

The Mbed serial terminal makes the following modifications to the default Miniterm.
* Custom help menu text
* Constrained set of menu keys
* CTRL-H to show help
* CTRL-B sends serial break to the target

To start the terminal clients should call the "run" function, this is the entry point to the module.
"""
from typing import Any

from serial import Serial
from serial.tools.miniterm import Miniterm


def run(port: str, baud: int, echo: bool = True) -> None:
    """Run the serial terminal.

    This function is blocking as it waits for the terminal thread to finish executing before returning.

    Args:
        port: The serial port to open a terminal on.
        baud: Serial baud rate.
        echo: Echo user input back to the console.
    """
    term = SerialTerminal(Serial(port=port, baudrate=str(baud)), echo=echo)
    term.start()

    try:
        term.join(True)
    except KeyboardInterrupt:
        pass
    finally:
        term.join()
        term.close()


class SerialTerminal(Miniterm):
    """An implementation of Miniterm that implements the additional Mbed terminal functionality.

    Overrides the `writer` method to implement modified menu key handling behaviour.
    Overrides the Miniterm::get_help_text method to return the Mbed custom help text.
    Adds a `reset` method so users can send a reset signal to the device.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Set the rx/tx encoding and special characters."""
        super().__init__(*args, **kwargs)
        self.exit_character = CTRL_C
        self.menu_character = CTRL_T
        self.reset_character = CTRL_B
        self.help_character = CTRL_H
        self.set_rx_encoding("UTF-8")
        self.set_tx_encoding("UTF-8")

    def reset(self) -> None:
        """Send a reset signal."""
        self.serial.sendBreak()

    def get_help_text(self) -> str:
        """Return the text displayed when the user requests help."""
        return HELP_TEXT

    def writer(self) -> None:
        """Implements terminal behaviour."""
        menu_active = False
        while self.alive:
            try:
                input_key = self.console.getkey()
            except KeyboardInterrupt:
                input_key = self.exit_character

            if (menu_active and input_key in VALID_MENU_KEYS) or (input_key == self.help_character):
                self.handle_menu_key(input_key)
                menu_active = False

            elif input_key == self.menu_character:
                menu_active = True

            elif input_key == self.reset_character:
                self.reset()

            elif input_key == self.exit_character:
                self.stop()
                break

            else:
                self._write_transformed_char(input_key)

                if self.echo:
                    self._echo_transformed_char(input_key)

    def _write_transformed_char(self, text: str) -> None:
        for transformation in self.tx_transformations:
            text = transformation.tx(text)

        self.serial.write(self.tx_encoder.encode(text))

    def _echo_transformed_char(self, text: str) -> None:
        for transformation in self.tx_transformations:
            text = transformation.echo(text)

        self.console.write(text)


CTRL_B = "\x02"
CTRL_C = "\x03"
CTRL_H = "\x08"
CTRL_T = "\x14"
VALID_MENU_KEYS = ("p", "b", "\t", "\x01", "\x04", "\x05", "\x06", "\x0c", CTRL_C, CTRL_T)
HELP_TEXT = """--- Mbed Serial Terminal
--- Based on miniterm from pySerial
---
--- Ctrl+b     Send Break (reset target)
--- Ctrl+c     Exit terminal
--- Ctrl+h     Help
--- Ctrl+t     Menu escape key, followed by:
---    p       Change COM port
---    b       Change baudrate
---    Tab     Show detailed terminal info
---    Ctrl+a  Change encoding (default UTF-8)
---    Ctrl+f  Edit filters
---    Ctrl+e  Toggle local echo
---    Ctrl+l  Toggle EOL
---    Ctrl+r  Toggle RTS
---    Ctrl+d  Toggle DTR
---    Ctrl+c  Send control character to remote
---    Ctrl+t  Send control character to remote
"""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import pathlib
from unittest import TestCase, mock

from tests.devices.factories import CandidateDeviceFactory
from mbed_tools.devices._internal.candidate_device import CandidateDevice
from mbed_tools.devices._internal.darwin import system_profiler, diskutil, ioreg
from mbed_tools.devices._internal.darwin.device_detector import (
    DarwinDeviceDetector,
    InvalidCandidateDeviceDataError,
    _assemble_candidate_data,
    _build_candidate,
    _build_ioreg_device_name,
    _get_mount_points,
    _get_serial_port,
)


@mock.patch("mbed_tools.devices._internal.darwin.device_detector._build_candidate")
@mock.patch("mbed_tools.devices._internal.darwin.device_detector.system_profiler", spec_set=system_profiler)
class TestDarwinDeviceDetector(TestCase):
    def test_find_candidates_successful_build_yields_candidate(self, system_profiler, _build_candidate):
        device_data = {"some": "data"}
        system_profiler.get_end_usb_devices_data.return_value = [device_data]
        candidate = CandidateDeviceFactory()
        _build_candidate.return_value = candidate
        self.assertEqual(DarwinDeviceDetector().find_candidates(), [candidate])
        _build_candidate.assert_called_with(device_data)

    def test_find_candidates_does_not_yield_failed_candidate_builds(self, system_profiler, _build_candidate):
        device_data = {"other": "data"}
        system_profiler.get_end_usb_devices_data.return_value = [device_data]
        _build_candidate.side_effect = InvalidCandidateDeviceDataError
        self.assertEqual(DarwinDeviceDetector().find_candidates(), [])
        _build_candidate.assert_called_with(device_data)


@mock.patch("mbed_tools.devices._internal.darwin.device_detector._assemble_candidate_data")
class TestBuildCandidateDevice(TestCase):
    def test_builds_candidate_using_assembled_data(self, _assemble_candidate_data):
        device_data = {
            "vendor_id": "0xff",
            "product_id": "0x24",
            "serial_number": "123456",
            "mount_points": ["/Volumes/A"],
            "serial_port": "port-1",
        }
        _assemble_candidate_data.return_value = device_data

        self.assertEqual(
            _build_candidate(device_data), CandidateDevice(**device_data),
        )

    @mock.patch("mbed_tools.devices._internal.darwin.device_detector.CandidateDevice")
    def test_raises_if_candidate_cannot_be_built(self, CandidateDevice, _assemble_candidate_data):
        CandidateDevice.side_effect = ValueError
        with self.assertRaises(InvalidCandidateDeviceDataError):
            _build_candidate({})


@mock.patch("mbed_tools.devices._internal.darwin.device_detector._get_serial_port")
@mock.patch("mbed_tools.devices._internal.darwin.device_detector._get_mount_points")
class TestAssembleCandidateDeviceData(TestCase):
    def test_glues_device_data_from_various_sources(self, _get_mount_points, _get_serial_port):
        device_data = {
            "vendor_id": "0xff",
            "product_id": "0x24",
            "serial_num": "123456",
        }
        _get_serial_port.return_value = "port-1"
        _get_mount_points.return_value = ["/Volumes/A"]

        self.assertEqual(
            _assemble_candidate_data(device_data),
            {
                "vendor_id": device_data.get("vendor_id"),
                "product_id": device_data.get("product_id"),
                "serial_number": device_data.get("serial_num"),
                "serial_port": _get_serial_port.return_value,
                "mount_points": _get_mount_points.return_value,
            },
        )

    def test_formats_vendor_id_containing_vendor_name(self, _get_mount_points, _get_serial_port):
        device_data = {"vendor_id": "0x12  (SomeVendor)"}
        result = _assemble_candidate_data(device_data)
        self.assertEqual(result["vendor_id"], "0x12")


class TestGetMountPoints(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.device_detector.diskutil", spec_set=diskutil)
    def test_maps_storage_identifiers_to_mount_points(self, diskutil):
        device_data = {"Media": [{"bsd_name": "disk1"}, {"bsd_name": "disk2"}]}
        diskutil.get_mount_point.side_effect = ["/Volumes/Disk1", "/Volumes/Disk2"]

        self.assertEqual(
            _get_mount_points(device_data), (pathlib.Path("/Volumes/Disk1"), pathlib.Path("/Volumes/Disk2"))
        )
        diskutil.get_mount_point.assert_has_calls([mock.call("disk1"), mock.call("disk2")])


class TestGetSerialPort(TestCase):
    @mock.patch("mbed_tools.devices._internal.darwin.device_detector.ioreg", spec_set=ioreg)
    def test_returns_retrieved_io_dialin_device(self, ioreg):
        """Given enough data, it constructs an ioreg device name and fetches serial port information."""
        device_data = {
            "location_id": "0x12345 / 2",
            "_name": "SomeDevice",
        }
        serial_port = "/dev/tty.usb1234"
        ioreg.get_io_dialin_device.return_value = serial_port
        ioreg_device_name = _build_ioreg_device_name(
            device_name=device_data["_name"], location_id=device_data["location_id"],
        )

        self.assertEqual(_get_serial_port(device_data), serial_port)
        ioreg.get_io_dialin_device.assert_called_once_with(ioreg_device_name)

    @mock.patch("mbed_tools.devices._internal.darwin.device_detector.ioreg", spec_set=ioreg)
    def test_returns_none_when_cant_determine_ioreg_name(self, ioreg):
        self.assertIsNone(_get_serial_port({}))


class TestBuildIoregDeviceName(TestCase):
    def test_builds_ioreg_device_name_from_system_profiler_data(self):
        self.assertEqual(
            _build_ioreg_device_name(device_name="VeryNiceDevice Really", location_id="0x14420000 / 2",),
            "VeryNiceDevice Really@14420000",
        )



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Tests for `mbed_tools.targets.boards`."""

import json
from dataclasses import asdict
from unittest import mock, TestCase

from mbed_tools.targets import Board
from mbed_tools.targets.boards import Boards
from mbed_tools.targets.exceptions import UnknownBoard
from tests.targets.factories import make_dummy_internal_board_data


@mock.patch("mbed_tools.targets._internal.board_database.get_online_board_data")
class TestBoards(TestCase):
    """Tests for the class `Boards`."""

    def test_iteration_is_repeatable(self, mocked_get_board_data):
        """Test Boards is an iterable and not an exhaustible iterator."""
        fake_board_data = make_dummy_internal_board_data()
        mocked_get_board_data.return_value = fake_board_data

        boards = Boards.from_online_database()
        tgts_a = [b for b in boards]
        tgts_b = [b for b in boards]

        self.assertEqual(tgts_a, tgts_b, "The lists are equal as boards was not exhausted on the first pass.")

    def test_board_found_in_boards_membership_test(self, mocked_get_board_data):
        """Tests the __contains__ method was implemented correctly."""
        board_data = make_dummy_internal_board_data()
        mocked_get_board_data.return_value = board_data

        boards = Boards.from_online_database()
        board, *_ = boards

        self.assertIn(board, boards)

    def test_membership_test_returns_false_for_non_board(self, mocked_get_board_data):
        """Tests the __contains__ method was implemented correctly."""
        board_data = make_dummy_internal_board_data()
        mocked_get_board_data.return_value = board_data

        boards = Boards.from_online_database()

        self.assertFalse("a" in boards)

    def test_len_boards(self, mocked_get_board_data):
        """Test len(Boards()) matches len(board_database)."""
        board_data = make_dummy_internal_board_data()
        mocked_get_board_data.return_value = board_data

        self.assertEqual(len(Boards.from_online_database()), len(board_data))

    def test_get_board_success(self, mocked_get_board_data):
        """Check a Board can be looked up by arbitrary parameters."""
        fake_board_data = [
            {"attributes": {"product_code": "0300"}},
            {"attributes": {"product_code": "0200"}},
            {"attributes": {"product_code": "0100"}},
        ]
        mocked_get_board_data.return_value = fake_board_data

        boards = Boards.from_online_database()
        board = boards.get_board(lambda b: b.product_code == "0100")

        self.assertEqual(board.product_code, "0100", "Board's product code should match the given product code.")

    def test_get_board_failure(self, mocked_get_board_data):
        """Check Boards handles queries without a match."""
        mocked_get_board_data.return_value = []

        boards = Boards.from_online_database()

        with self.assertRaises(UnknownBoard):
            boards.get_board(lambda b: b.product_code == "unknown")

    @mock.patch("mbed_tools.targets._internal.board_database.get_offline_board_data")
    def test_json_dump_from_raw_and_filtered_data(self, mocked_get_offline_board_data, mocked_get_online_board_data):
        raw_board_data = [
            {"attributes": {"product_code": "0200", "board": "test"}},
            {"attributes": {"product_code": "0100", "board": "test2"}},
        ]
        mocked_get_online_board_data.return_value = raw_board_data

        boards = [Board.from_online_board_entry(b) for b in raw_board_data]
        filtered_board_data = [asdict(board) for board in boards]
        mocked_get_offline_board_data.return_value = filtered_board_data

        # Boards.from_online_database handles "raw" board entries from the online db
        boards = Boards.from_online_database()
        json_str_from_raw = boards.json_dump()
        t1_raw, t2_raw = boards

        # Boards.from_offline_database expects the data to have been "filtered" through the Boards interface
        offline_boards = Boards.from_offline_database()
        json_str_from_filtered = offline_boards.json_dump()
        t1_filt, t2_filt = offline_boards

        self.assertEqual(
            json_str_from_raw,
            json.dumps([asdict(t1_raw), asdict(t2_raw)], indent=4),
            "JSON string should match serialised board __dict__.",
        )

        self.assertEqual(json_str_from_filtered, json.dumps([t1_filt.__dict__, t2_filt.__dict__], indent=4))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a Logical disk."""

from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor


class LogicalDiskMsdnDefinition(NamedTuple):
    """Msdn definition of a logical disk.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/cim-logicaldisk
    """

    Access: int
    Availability: int
    BlockSize: int
    Caption: str
    ConfigManagerErrorCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    Description: str
    DeviceID: str
    ErrorCleared: bool
    ErrorDescription: str
    ErrorMethodology: str
    FreeSpace: int
    InstallDate: int
    LastErrorCode: int
    Name: str
    NumberOfBlocks: int
    PNPDeviceID: str
    PowerManagementCapabilities: list
    PowerManagementSupported: bool
    Purpose: str
    Size: int
    Status: str
    StatusInfo: int
    SystemCreationClassName: str
    SystemName: str


class LogicalDisk(ComponentDescriptor):
    """Logical disk as defined in Windows API.

    See https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/cim-logicaldisk
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(LogicalDiskMsdnDefinition, win32_class_name="CIM_LogicalDisk")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Environment options for `mbed-targets`.

All the env configuration options can be set either via environment variables or using a `.env` file
containing the variable definitions as follows:

```
VARIABLE=value
```

Environment variables take precendence, meaning the values set in the file will be overriden
by any values previously set in your environment.

.. WARNING::
   Do not upload `.env` files containing private tokens to version control! If you use this package
   as a dependency of your project, please ensure to include the `.env` in your `.gitignore`.
"""
import os

import dotenv
import pdoc

dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))


class Env:
    """Provides access to environment variables.

    Ensures variables are reloaded when environment changes during runtime.
    Additionally allows to expose documented instance variables in pdoc
    generated output.
    """

    @property
    def MBED_API_AUTH_TOKEN(self) -> str:
        """Token to use when accessing online API.

        Mbed Targets uses the online mbed board database at os.mbed.com as its data source.
        A snapshot of the board database is shipped with the package, for faster lookup of known
        boards. Only public boards are stored in the database snapshot. If you are fetching data
        for a private board, mbed-targets will need to contact the online database.

        To fetch data about private boards from the online database, the user must have an account
        on os.mbed.com and be member of a vendor team that has permissions to see the private board.
        An authentication token for the team member must be provided in an environment variable named
        `MBED_API_AUTH_TOKEN`.
        """
        return os.getenv("MBED_API_AUTH_TOKEN", "")

    @property
    def MBED_DATABASE_MODE(self) -> str:
        """Database mode to use when retrieving board data.

        Mbed Targets supports an online and offline mode, which controls where to look up the board database.

        The board lookup can be from either the online or offline database, depending
        on the value of an environment variable called `MBED_DATABASE_MODE`.

        The mode can be set to one of the following:

        - `AUTO`: the offline database is searched first, if the board isn't found the online database is searched.
        - `ONLINE`: the online database is always used.
        - `OFFLINE`: the offline database is always used.

        If `MBED_DATABASE_MODE` is not set, it defaults to `AUTO`.
        """
        return os.getenv("MBED_DATABASE_MODE", "AUTO")


env = Env()
"""Instance of `Env` class."""

env_variables = pdoc.Class("Env", pdoc.Module("mbed_tools.targets.env"), Env).instance_variables()



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
import unittest
from tests.devices.markers import windows_only


@windows_only
class TestDiskDataAggregrator(unittest.TestCase):
    def test_data_aggregation(self):
        from mbed_tools.devices._internal.windows.disk_drive import DiskDrive
        from mbed_tools.devices._internal.windows.disk_partition import DiskPartition
        from mbed_tools.devices._internal.windows.logical_disk import LogicalDisk
        from mbed_tools.devices._internal.windows.volume_set import VolumeInformation, DriveType
        from mbed_tools.devices._internal.windows.disk_aggregation import (
            AggregatedDiskData,
            DiskDataAggregator,
            WindowsUID,
        )

        disk1 = DiskDrive()
        disk1.set_data_values(
            dict(
                Availability=None,
                BytesPerSector=512,
                Capabilities=(3, 4, 7),
                CapabilityDescriptions=("Random Access", "Supports Writing", "Supports Removable Media"),
                Caption="MBED VFS USB Device",
                CompressionMethod=None,
                ConfigManagerErrorCode=0,
                ConfigManagerUserConfig=False,
                CreationClassName="Win32_DiskDrive",
                DefaultBlockSize=None,
                Description="Disk drive",
                DeviceID="\\\\.\\PHYSICALDRIVE1",
                ErrorCleared=None,
                ErrorDescription=None,
                ErrorMethodology=None,
                FirmwareRevision="0.1 ",
                Index=1,
                InstallDate=None,
                InterfaceType="USB",
                LastErrorCode=None,
                Manufacturer="(Standard disk drives)",
                MaxBlockSize=None,
                MaxMediaSize=None,
                MediaLoaded=True,
                MediaType="Removable Media",
                MinBlockSize=None,
                Model="MBED VFS USB Device",
                Name="\\\\.\\PHYSICALDRIVE1",
                NeedsCleaning=None,
                NumberOfMediaSupported=None,
                Partitions=1,
                PNPDeviceID="USBSTOR\\DISK&VEN_MBED&PROD_VFS&REV_0.1\\4454646",
                PowerManagementCapabilities=None,
                PowerManagementSupported=None,
                SCSIBus=0,
                SCSILogicalUnit=0,
                SCSIPort=0,
                SCSITargetId=0,
                SectorsPerTrack=63,
                SerialNumber="0240000034544e45001a00018aa900292011000097969900",
                Signature=1,
                Size="65802240",
                Status="OK",
                StatusInfo=None,
                SystemCreationClassName="Win32_ComputerSystem",
                SystemName="E112100",
                TotalCylinders="8",
                TotalHeads=255,
                TotalSectors="128520",
                TotalTracks="2040",
                TracksPerCylinder=255,
            )
        )
        partition1 = DiskPartition()
        partition1.set_data_values(
            dict(
                AdditionalAvailability=None,
                Availability=None,
                PowerManagementCapabilities=None,
                IdentifyingDescriptions=None,
                MaxQuiesceTime=None,
                OtherIdentifyingInfo=None,
                StatusInfo=None,
                PowerOnHours=None,
                TotalPowerOnHours=None,
                Access=None,
                BlockSize="512",
                Bootable=False,
                BootPartition=False,
                Caption="Disk #1, Partition #0",
                ConfigManagerErrorCode=None,
                ConfigManagerUserConfig=None,
                CreationClassName="Win32_DiskPartition",
                Description="16-bit FAT",
                DeviceID="Disk #1, Partition #0",
                DiskIndex=1,
                ErrorCleared=None,
                ErrorDescription=None,
                ErrorMethodology=None,
                HiddenSectors=None,
                Index=0,
                InstallDate=None,
                LastErrorCode=None,
                Name="Disk #1, Partition #0",
                NumberOfBlocks="131200",
                PNPDeviceID=None,
                PowerManagementSupported=None,
                PrimaryPartition=True,
                Purpose=None,
                RewritePartition=None,
                Size="67174400",
                StartingOffset="0",
                Status=None,
                SystemCreationClassName="Win32_ComputerSystem",
                SystemName="E112100",
                Type="16-bit FAT",
            )
        )
        logical_disk1 = LogicalDisk()
        logical_disk1.set_data_values(
            dict(
                Access=0,
                Availability=None,
                BlockSize=None,
                Caption="F:",
                ConfigManagerErrorCode=None,
                ConfigManagerUserConfig=None,
                CreationClassName="Win32_LogicalDisk",
                Description="Removable Disk",
                DeviceID="F:",
                ErrorCleared=None,
                ErrorDescription=None,
                ErrorMethodology=None,
                FreeSpace="67096576",
                InstallDate=None,
                LastErrorCode=None,
                Name="F:",
                NumberOfBlocks=None,
                PNPDeviceID=None,
                PowerManagementCapabilities=None,
                PowerManagementSupported=None,
                Purpose=None,
                Size="67104768",
                Status=None,
                StatusInfo=None,
                SystemCreationClassName="Win32_ComputerSystem",
                SystemName="E112100",
            )
        )
        disk_drives = {1: disk1}
        disk_partitions = {"Disk #1, Partition #0": partition1}
        logical_partition_relationships = {"F:": "Disk #1, Partition #0"}
        volume_information = VolumeInformation(
            Name="DAPLINK",
            SerialNumber=654449012,
            MaxComponentLengthOfAFileName=255,
            SysFlags=131590,
            FileSystem="FAT",
            UniqueName="\\\\?\\Volume{d0613192-49b4-11ea-99e5-c85b76dfd333}\\",
            DriveType=DriveType.DRIVE_REMOVABLE,
        )

        # TEST aggregation
        expected_aggregated_object = AggregatedDiskData()
        expected_aggregated_object.set_data_values(
            dict(
                uid=WindowsUID(uid="4454646", raw_uid=None, serial_number=None),
                label="F:",
                description="Removable Disk",
                free_space="67096576",
                size="67104768",
                partition_name="Disk #1, Partition #0",
                partition_type="16-bit FAT",
                volume_information=volume_information,
                caption="MBED VFS USB Device",
                physical_disk_name="\\\\.\\PHYSICALDRIVE1",
                model="MBED VFS USB Device",
                interface_type="USB",
                media_type="Removable Media",
                manufacturer="(Standard disk drives)",
                serial_number="0240000034544e45001a00018aa900292011000097969900",
                status="OK",
                pnp_device_id="USBSTOR\\DISK&VEN_MBED&PROD_VFS&REV_0.1\\4454646",
            )
        )

        self.assertTupleEqual(
            DiskDataAggregator(
                physical_disks=disk_drives,
                partition_disks=disk_partitions,
                logical_partition_relationships=logical_partition_relationships,
                lookup_volume_information=lambda ld: volume_information,
            )
            .aggregate(logical_disk1)
            .to_tuple(),
            expected_aggregated_object.to_tuple(),
        )

    def test_system_disk_data(self):
        # On Windows, C: is the default primary drive and hence we should always get information about it.
        from mbed_tools.devices._internal.windows.disk_aggregation import SystemDiskInformation
        from mbed_tools.devices._internal.windows.system_data_loader import SystemDataLoader

        disks = SystemDiskInformation(SystemDataLoader())
        c_data = disks.get_disk_information_by_label("c:")
        self.assertFalse(c_data.is_undefined)
        uid = c_data.uid
        self.assertIsNotNone(uid)
        labels = [d.label.lower() for d in disks.get_disk_information(uid)]
        self.assertTrue("c:" in labels)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Configuration assembly algorithm."""
import itertools

from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional, Set

from mbed_tools.build._internal.config.config import Config
from mbed_tools.build._internal.config import source
from mbed_tools.build._internal.find_files import LabelFilter, RequiresFilter, filter_files, find_files


def assemble_config(target_attributes: dict, search_paths: Iterable[Path], mbed_app_file: Optional[Path]) -> Config:
    """Assemble config for given target and program directory.

    Mbed library and application specific config parameters are parsed from mbed_lib.json and mbed_app.json files
    located in the project source tree.
    The config files contain sets of "labels" which correspond to directory names in the mbed-os source tree. These
    labels are used to determine which mbed_lib.json files to include in the final configuration.

    The mbed_app.json config overrides must be applied after all the mbed_lib.json files have been parsed.
    Unfortunately, mbed_app.json may also contain filter labels to tell us which mbed libs we're depending on.
    This means we have to collect the filter labels from mbed_app.json before parsing any other config files.
    Then we parse all the required mbed_lib config and finally apply the app overrides.

    Args:
        target_attributes: Mapping of target specific config parameters.
        search_paths: Iterable of paths to search for mbed_lib.json files.
        mbed_app_file: The path to mbed_app.json. This can be None.
    """
    mbed_lib_files = list(
        set(itertools.chain.from_iterable(find_files("mbed_lib.json", path) for path in search_paths))
    )
    return _assemble_config_from_sources(target_attributes, mbed_lib_files, mbed_app_file)


def _assemble_config_from_sources(
    target_attributes: dict, mbed_lib_files: List[Path], mbed_app_file: Optional[Path] = None
) -> Config:
    config = Config(source.prepare(target_attributes, source_name="target"))
    previous_filter_data = None
    app_data = None
    if mbed_app_file:
        # We need to obtain the file filter data from mbed_app.json so we can select the correct set of mbed_lib.json
        # files to include in the config. We don't want to update the config object with all of the app settings yet
        # as we won't be able to apply overrides correctly until all relevant mbed_lib.json files have been parsed.
        app_data = source.from_file(
            mbed_app_file, default_name="app", target_filters=FileFilterData.from_config(config).labels
        )
        _get_app_filter_labels(app_data, config)

    current_filter_data = FileFilterData.from_config(config)
    while previous_filter_data != current_filter_data:
        filtered_files = _filter_files(mbed_lib_files, current_filter_data)
        for config_file in filtered_files:
            config.update(source.from_file(config_file, target_filters=current_filter_data.labels))
            # Remove any mbed_lib files we've already visited from the list so we don't parse them multiple times.
            mbed_lib_files.remove(config_file)

        previous_filter_data = current_filter_data
        current_filter_data = FileFilterData.from_config(config)

    # Apply mbed_app.json data last so config parameters are overriden in the correct order.
    if app_data:
        config.update(app_data)

    return config


def _get_app_filter_labels(mbed_app_data: dict, config: Config) -> None:
    requires = mbed_app_data.get("requires")
    if requires:
        config["requires"] = requires

    config.update(_get_file_filter_overrides(mbed_app_data))


def _get_file_filter_overrides(mbed_app_data: dict) -> dict:
    return {
        "overrides": [
            override
            for override in mbed_app_data.get("overrides", [])
            if override.modifier or override.name == "requires"
        ]
    }


@dataclass(frozen=True)
class FileFilterData:
    """Data used to navigate mbed-os directories for config files."""

    labels: Set[str]
    features: Set[str]
    components: Set[str]
    requires: Set[str]

    @classmethod
    def from_config(cls, config: Config) -> "FileFilterData":
        """Extract file filters from a Config object."""
        return cls(
            labels=config.get("labels", set()) | config.get("extra_labels", set()),
            features=set(config.get("features", set())),
            components=set(config.get("components", set())),
            requires=set(config.get("requires", set())),
        )


def _filter_files(files: Iterable[Path], filter_data: FileFilterData) -> Iterable[Path]:
    filters = (
        LabelFilter("TARGET", filter_data.labels),
        LabelFilter("FEATURE", filter_data.features),
        LabelFilter("COMPONENT", filter_data.components),
        RequiresFilter(filter_data.requires),
    )
    return filter_files(files, filters)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from tests.devices.markers import windows_only
from unittest import TestCase
import random


@windows_only
class TestSerialPort(TestCase):
    def test_retrieve_port_name(self):
        from mbed_tools.devices._internal.windows.serial_port import parse_caption
        from mbed_tools.devices._internal.windows.component_descriptor import UNKNOWN_VALUE

        self.assertEqual(UNKNOWN_VALUE, parse_caption(UNKNOWN_VALUE))
        self.assertEqual("COM13", parse_caption("Serial Port for Barcode Scanner (COM13)"))
        port_name = f"COM{random.choice(range(0, 1000))}"
        self.assertEqual(port_name, parse_caption(f"mbed Serial Port ({port_name})"))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Defines a USB hub."""

from typing import NamedTuple, cast

from mbed_tools.devices._internal.windows.component_descriptor import ComponentDescriptor


class UsbHubMsdnDefinition(NamedTuple):
    """Msdn definition of a Usb hub.

    See https://docs.microsoft.com/en-gb/previous-versions/windows/desktop/cimwin32a/win32-usbhub?redirectedfrom=MSDN
    See also https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/cim-usbdevice
    """

    Availability: int
    Caption: str
    ClassCode: int
    ConfigManagerUserConfig: bool
    CreationClassName: str
    CurrentAlternateSettings: list
    CurrentConfigValue: int
    Description: str
    ErrorCleared: bool
    ErrorDescription: str
    GangSwitched: bool
    InstallDate: int
    LastErrorCode: int
    NumberOfConfigs: int
    NumberOfPorts: int
    PNPDeviceID: str
    PowerManagementCapabilities: list
    PowerManagementSupported: bool
    ProtocolCode: int
    Status: str
    StatusInfo: int
    SubclassCode: int
    SystemCreationClassName: str
    SystemName: str
    USBVersion: int
    ConfigManagerErrorCode: int
    DeviceID: str
    Name: str


class UsbHub(ComponentDescriptor):
    """USB Hub as defined in Windows API.

    See https://docs.microsoft.com/en-gb/previous-versions/windows/desktop/cimwin32a/win32-usbhub?redirectedfrom=MSDN
    Seems similar to https://docs.microsoft.com/en-us/windows/win32/cimwin32prov/cim-usbhub
    """

    def __init__(self) -> None:
        """Initialiser."""
        super().__init__(UsbHubMsdnDefinition, win32_class_name="Win32_USBHub")

    @property
    def component_id(self) -> str:
        """Returns the device id field."""
        return cast(str, self.get("DeviceID"))

    @property
    def pnp_id(self) -> str:
        """Returns the plug and play id field."""
        return cast(str, self.get("PNPDeviceID"))



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Exceptions raised by mbed tools."""


class ToolsError(Exception):
    """Base class for tools errors."""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""Public exceptions exposed by the package."""

from mbed_tools.lib.exceptions import ToolsError


class MbedTargetsError(ToolsError):
    """Base exception for mbed-targets."""


class TargetError(ToolsError):
    """Target definition cannot be retrieved."""


class UnknownBoard(MbedTargetsError):
    """Requested board was not found."""


class UnsupportedMode(MbedTargetsError):
    """The Database Mode is unsupported."""


class BoardDatabaseError(MbedTargetsError):
    """Failed to get the board data from the database."""



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
"""mbed_tools command line interface."""

from mbed_tools.cli.main import cli, LOGGER



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#

import pathlib
import tempfile

from unittest import TestCase, mock

from mbed_tools.build.flash import flash_binary, _build_binary_file_path, _flash_dev
from mbed_tools.build.exceptions import BinaryFileNotFoundError

from tests.build.factories import DeviceFactory


@mock.patch("mbed_tools.build.flash._build_binary_file_path")
@mock.patch("mbed_tools.build.flash._flash_dev")
class TestFlashBinary(TestCase):
    def test_check_flashing(self, _flash_dev, _build_binary_file_path):
        test_device = DeviceFactory()

        _flash_dev.return_value = True

        with tempfile.TemporaryDirectory() as tmpDir:
            base_dir = pathlib.Path(tmpDir)
            build_dir = base_dir / "cmake_build"
            build_dir.mkdir()
            bin_file = base_dir.name + ".bin"
            bin_file = build_dir / bin_file
            bin_file.touch()
            _build_binary_file_path.return_value = bin_file

            flash_binary(test_device.mount_points[0].resolve(), base_dir, build_dir, "TEST", False)

            _build_binary_file_path.assert_called_once_with(base_dir, build_dir, False)
            _flash_dev.assert_called_once_with(test_device.mount_points[0].resolve(), bin_file)


class TestBuildBinFilePath(TestCase):
    def test_build_bin_file_path(self):
        with tempfile.TemporaryDirectory() as tmpDir:
            base_dir = pathlib.Path(tmpDir)
            build_dir = base_dir / "cmake_build"
            build_dir.mkdir()
            bin_file = base_dir.name + ".bin"
            bin_file = build_dir / bin_file
            bin_file.touch()

            self.assertEqual(_build_binary_file_path(base_dir, build_dir, False), bin_file)

    def test_build_hex_file_path(self):
        with tempfile.TemporaryDirectory() as tmpDir:
            base_dir = pathlib.Path(tmpDir)
            build_dir = base_dir / "cmake_build"
            build_dir.mkdir()
            bin_file = base_dir.name + ".hex"
            bin_file = build_dir / bin_file
            bin_file.touch()

            self.assertEqual(_build_binary_file_path(base_dir, build_dir, True), bin_file)

    def test_missing_binary_file(self):
        with tempfile.TemporaryDirectory() as tmpDir:
            base_dir = pathlib.Path(tmpDir)
            build_dir = base_dir / "cmake_build"
            build_dir.mkdir()

            with self.assertRaises(BinaryFileNotFoundError):
                _build_binary_file_path(base_dir, build_dir, False)


@mock.patch("mbed_tools.build.flash.shutil.copy")
class TestCopyToDevice(TestCase):
    def test_copy_to_device(self, copy):
        test_device = DeviceFactory()

        with tempfile.TemporaryDirectory() as tmpDir:
            base_dir = pathlib.Path(tmpDir)
            build_dir = base_dir / "cmake_build"
            build_dir.mkdir()
            bin_file = base_dir.name + ".bin"
            bin_file = build_dir / bin_file
            bin_file.touch()
            _flash_dev(test_device.mount_points[0].resolve(), bin_file)

            copy.assert_called_once_with(bin_file, test_device.mount_points[0].resolve(), follow_symlinks=False)



#
# Copyright (c) 2020-2021 Arm Limited and Contributors. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
from mbed_tools.build._internal.config import source
from mbed_tools.build._internal.config.source import Override


class TestPrepareSource:
    def test_config_fields_from_target_are_namespaced(self):
        target = {
            "config": {
                "network-default-interface-type": {
                    "help": "Default network "
                    "interface type. "
                    "Typical options: null, "
                    "ETHERNET, WIFI, "
                    "CELLULAR, MESH",
                    "value": "ETHERNET",
                }
            }
        }

        conf = source.prepare(target, "target")

        config_setting, *_ = conf["config"]
        assert config_setting.namespace == "target"
        assert config_setting.name == "network-default-interface-type"

    def test_override_fields_from_target_are_namespaced(self):
        target = {"overrides": {"network-default-interface-type": "ETHERNET"}}

        conf = source.prepare(target, "target")

        network_override, *_ = conf["overrides"]
        assert network_override.namespace == "target" and network_override.name == "network-default-interface-type"

    def test_config_fields_from_lib_are_namespaced(self):
        lib = {
            "name": "library",
            "config": {
                "network-default-interface-type": {
                    "help": "Default network "
                    "interface type. "
                    "Typical options: null, "
                    "ETHERNET, WIFI, "
                    "CELLULAR, MESH",
                    "value": "ETHERNET",
                }
            },
        }

        conf = source.prepare(lib)

        config_setting, *_ = conf["config"]
        assert config_setting.namespace == "library"
        assert config_setting.name == "network-default-interface-type"

    def test_override_fields_from_lib_are_namespaced(self):
        lib = {
            "name": "lib",
            "target_overrides": {"*": {"network-default-interface-type": "ETHERNET", "target.device_has": ["k"]}},
        }

        conf = source.prepare(lib)

        network_override, device_has_override = conf["overrides"]
        assert network_override.namespace == "lib" and network_override.name == "network-default-interface-type"
        assert device_has_override.namespace == "target" and device_has_override.name == "device_has"

    def test_target_overrides_only_collected_for_valid_targets(self):
        lib = {
            "name": "lib",
            "target_overrides": {
                "*": {"target.macros": ["j"]},
                "VALID_TARGET": {"target.device_has": ["k"]},
                "FILTER_TARGET": {"network-default-interface-type": "ETHERNET"},
            },
        }
        expected_macro_override = Override(namespace="target", name="macros", value={"j"}, modifier=None)
        expected_device_has_override = Override(namespace="target", name="device_has", value={"k"}, modifier=None)

        conf = source.prepare(lib, target_filters=["VALID_TARGET"])

        macro_override, device_has_override, *others = conf["overrides"]
        assert macro_override == expected_macro_override
        assert device_has_override == expected_device_has_override
        assert others == []

    def test_cumulative_fields_parsed(self):
        lib = {
            "name": "lib",
            "target_overrides": {
                "*": {"macros_add": ["MAC"], "target.device_has_add": ["k"], "target.device_has_remove": ["j"]}
            },
        }
        expected_device_has_add = Override(namespace="target", name="device_has", modifier="add", value={"k"})
        expected_device_has_remove = Override(namespace="target", name="device_has", modifier="remove", value={"j"})
        expected_macros_add = Override(namespace="lib", name="macros", modifier="add", value={"MAC"})

        conf = source.prepare(lib)

        macros_add_override, device_has_add_override, device_has_remove_override = conf["overrides"]
        assert device_has_add_override == expected_device_has_add
        assert device_has_remove_override == expected_device_has_remove
        assert macros_add_override == expected_macros_add

    def test_converts_config_setting_value_lists_to_sets(self):
        lib = {
            "name": "library",
            "config": {"list-values": {"value": ["ETHERNET", "WIFI"]}},
            "sectors": [[0, 2048]],
            "header_info": [[0, 2048], ["bobbins"], ["magic"]],
        }

        conf = source.prepare(lib)

        assert conf["config"][0].value == {"ETHERNET", "WIFI"}
        assert conf["sectors"] == {0, 2048}
        assert conf["header_info"] == {0, 2048, "bobbins", "magic"}
